import keras as ks
from keras.backend import backend as backend_to_use
from kgcnn.layers.modules import Embedding
from kgcnn.layers.mlp import MLP
from kgcnn.models.utils import update_model_kwargs
from ._layers import MATAttentionHead, MATDistanceMatrix, MATReduceMask, MATGlobalPool, MATExpandMask

# Keep track of model version from commit date in literature.
# To be updated if model is changed in a significant way.
__model_version__ = "2023-12-08"

# Supported backends
__kgcnn_model_backend_supported__ = ["tensorflow", "torch", "jax"]
if backend_to_use() not in __kgcnn_model_backend_supported__:
    raise NotImplementedError("Backend '%s' for model 'MAT' is not supported." % backend_to_use())

# Implementation of MAT in `tf.keras` from paper:
# Molecule Attention Transformer
# Łukasz Maziarka, Tomasz Danel, Sławomir Mucha, Krzysztof Rataj, Jacek Tabor, Stanisław Jastrzębski
# https://arxiv.org/abs/2002.08264
# https://github.com/ardigen/MAT
# https://github.com/lucidrains/molecule-attention-transformer


model_default = {
    "name": "MAT",
    "inputs": [
        {"shape": (None,), "name": "node_number", "dtype": "int64"},
        {"shape": (None, 3), "name": "node_coordinates", "dtype": "float32"},
        {"shape": (None, None, 1), "name": "adjacency_matrix", "dtype": "float32"},
        {"shape": (None,), "name": "node_mask", "dtype": "bool"},
        {"shape": (None, None), "name": "adjacency_mask", "dtype": "bool"},
    ],
    "input_tensor_type": "padded",
    "input_embedding": None,
    "input_node_embedding": {"input_dim": 95, "output_dim": 64},
    "input_edge_embedding": {"input_dim": 95, "output_dim": 64},
    "max_atoms": None,
    "distance_matrix_kwargs": {"trafo": "exp"},
    "attention_kwargs": {"units": 8, "lambda_attention": 0.3, "lambda_distance": 0.3, "lambda_adjacency": None,
                         "dropout": 0.1, "add_identity": False},
    "feed_forward_kwargs": {"units": [32, 32, 32], "activation": ["relu", "relu", "linear"]},
    "embedding_units": 32,
    "depth": 5,
    "heads": 8,
    "merge_heads": "concat",
    "verbose": 10,
    "pooling_kwargs": {"pooling_method": "sum"},
    "output_embedding": "graph",
    "output_to_tensor": None,
    "output_mlp": {"use_bias": [True, True, True], "units": [32, 16, 1],
                   "activation": ["relu", "relu", "linear"]},
    "output_tensor_type": "padded"
}


@update_model_kwargs(model_default, update_recursive=0, deprecated=["input_embedding", "output_to_tensor"])
def make_model(name: str = None,
               inputs: list = None,
               input_embedding: dict = None,
               input_node_embedding: dict = None,
               input_tensor_type: str = None,
               input_edge_embedding: dict = None,
               distance_matrix_kwargs: dict = None,
               attention_kwargs: dict = None,
               max_atoms: int = None,
               feed_forward_kwargs:dict = None,
               embedding_units: int = None,
               depth: int = None,
               heads: int = None,
               merge_heads: str = None,
               verbose: int = None,
               pooling_kwargs: dict = None,
               output_embedding: str = None,
               output_to_tensor: bool = None,
               output_mlp: dict = None,
               output_tensor_type: str = None
               ):
    r"""Make `MAT <https://arxiv.org/pdf/2002.08264.pdf>`__ graph network via functional API.
    Default parameters can be found in :obj:`kgcnn.literature.MAT.model_default` .

    .. note::

        We added a linear layer to keep correct node embedding dimension.

    Inputs:
        list: `[node_attributes, node_coordinates, adjacency_matrix, node_mask, adjacency_mask]`

            - node_attributes (Tensor): Node attributes of shape `(batch, N, F)` or `(batch, N)`
              using an embedding layer.
            - node_coordinates (Tensor): Node (atomic) coordinates of shape `(batch, N, 3)`.
            - adjacency_matrix (Tensor): Edge attributes of shape `(batch, N, N, F)` or `(batch, N, N)`
              using an embedding layer.
            - node_mask (Tensor): Node mask of shape `(batch, N)`
            - adjacency_mask (Tensor): Adjacency mask of shape `(batch, N, N)`

    Outputs:
        Tensor: Graph embeddings of shape `(batch, L)` if :obj:`output_embedding="graph"`.

    Args:
        name (str): Name of the model. Should be "MAT".
        inputs (list): List of dictionaries unpacked in :obj:`keras.layers.Input`. Order must match model definition.
        input_tensor_type (str): Input tensor type. Only "padded" is valid for this implementation.
        input_node_embedding (dict): Dictionary of embedding arguments unpacked in :obj:`Embedding` layers.
        input_edge_embedding (dict): Dictionary of embedding arguments unpacked in :obj:`Embedding` layers.
        depth (int): Number of graph embedding units or depth of the network.
        verbose (int): Level for print information.
        distance_matrix_kwargs (dict): Dictionary of layer arguments unpacked in :obj:`MATDistanceMatrix`.
        attention_kwargs (dict): Dictionary of layer arguments unpacked in :obj:`MATDistanceMatrix`.
        feed_forward_kwargs (dict): Dictionary of layer arguments unpacked in feed forward :obj:`MLP`.
        embedding_units (int): Units for node embedding.
        heads (int): Number of attention heads
        merge_heads (str): How to merge head, using either 'sum' or 'concat'.
        pooling_kwargs (dict): Dictionary of layer arguments unpacked in :obj:`MATGlobalPool`.
        output_embedding (str): Main embedding task for graph network. Either "node", "edge" or "graph".
        output_to_tensor (bool): Whether to cast model output to :obj:`Tensor`.
        output_mlp (dict): Dictionary of layer arguments unpacked in the final classification :obj:`MLP` layer block.
            Defines number of model outputs and activation.
        output_tensor_type (str): Output tensor type. Only "padded" is valid for this implementation.

    Returns:
        :obj:`keras.models.Model`
    """
    assert input_tensor_type in ["padded", "mask", "masked"], "Only padded tensors are valid for this implementation."
    assert output_tensor_type in ["padded", "mask", "masked"], "Only padded tensors are valid for this implementation."
    # Make input
    node_input = ks.layers.Input(**inputs[0])
    xyz_input = ks.layers.Input(**inputs[1])
    adjacency_matrix = ks.layers.Input(**inputs[2])
    node_mask = ks.layers.Input(**inputs[3])
    adjacency_mask = ks.layers.Input(**inputs[4])

    use_edge_embedding = input_edge_embedding is not None
    use_node_embedding = input_node_embedding is not None
    # Embedding, if no feature dimension
    if use_node_embedding:
        n = Embedding(**input_node_embedding)(node_input)
    else:
        n = node_input
    if use_edge_embedding:
        adj = Embedding(**input_edge_embedding)(adjacency_matrix)
    else:
        adj = adjacency_matrix
    n_mask = node_mask
    adj_mask = adjacency_mask
    xyz = xyz_input
    n_mask = MATExpandMask(axis=-1)(n_mask)
    adj_mask = MATExpandMask(axis=-1)(adj_mask)

    # Cast to dense Tensor with padding for MAT.
    # Nodes must have feature dimension.
    dist, dist_mask = MATDistanceMatrix(**distance_matrix_kwargs)(xyz, mask=n_mask)

    # Check shapes
    # print(n.shape, dist.shape, adj.shape)
    # print(n_mask.shape, dist_mask.shape, adj_mask.shape)

    # Adjacency is derived from edge input. If edge input has no last dimension and no embedding is used, then adjacency
    # matrix will have shape (batch, max_atoms, max_atoms) and edge input should be ones or weights or bond degree.
    # Otherwise, adjacency bears feature expanded from edge attributes of shape (batch, max_atoms, max_atoms, features).
    has_edge_dim = len(inputs[2]["shape"]) >= 3 or len(inputs[2]["shape"]) < 3 and use_edge_embedding

    if has_edge_dim:
        # Assume that feature-wise attention is not desired for adjacency, reduce to single value.
        adj = ks.layers.Dense(1, use_bias=False)(adj)
    else:
        # Make sure that shape is (batch, max_atoms, max_atoms, 1).
        adj = MATExpandMask(axis=-1)(adj)

    # Repeat for depth.
    h_mask = n_mask
    h = ks.layers.Dense(units=embedding_units, use_bias=False)(n)  # Assert correct feature dimension for skip.
    for _ in range(depth):
        # 1. Norm + Attention + Residual
        hn = ks.layers.LayerNormalization()(h)
        hs = [
            MATAttentionHead(**attention_kwargs)(
                [hn, dist, adj],
                mask=[n_mask, dist_mask, adj_mask]
            )
            for _ in range(heads)
        ]
        if merge_heads in ["add", "sum", "reduce_sum"]:
            hu = ks.layers.Add()(hs)
            hu = ks.layers.Dense(units=embedding_units, use_bias=False)(hu)
        else:
            hu = ks.layers.Concatenate(axis=-1)(hs)
            hu = ks.layers.Dense(units=embedding_units, use_bias=False)(hu)
        h = ks.layers.Add()([h, hu])

        # 2. Norm + MLP + Residual
        hn = ks.layers.LayerNormalization()(h)
        hu = MLP(**feed_forward_kwargs)(hn)
        hu = ks.layers.Dense(units=embedding_units, use_bias=False)(hu)
        hu = ks.layers.Multiply()([hu, h_mask])
        h = ks.layers.Add()([h, hu])

    # pooling output
    out = h
    out_mask = h_mask
    out = ks.layers.LayerNormalization()(out)
    if output_embedding == 'graph':
        out = ks.layers.Multiply()([out, out_mask])
        out = MATGlobalPool(**pooling_kwargs)(out, mask=out_mask)
        # final prediction MLP for the output!
        out = MLP(**output_mlp)(out)
    elif output_embedding == 'node':
        out = MLP(**output_mlp)(out)
        out = ks.layers.Multiply()([out, out_mask])
    else:
        raise ValueError("Unsupported graph embedding for mode `MAT` .")

    model = ks.models.Model(
        inputs=[node_input, xyz_input, adjacency_matrix, node_mask, adjacency_mask],
        outputs=out,
        name=name
    )
    model.__kgcnn_model_version__ = __model_version__
    return model



from keras import Layer
from keras import ops


class ExplanationSparsityRegularization(Layer):

    def __init__(self,
                 factor: float = 1.0,
                 **kwargs):
        super(ExplanationSparsityRegularization, self).__init__(**kwargs)
        self.factor = factor

    def build(self, input_shape):
        super(ExplanationSparsityRegularization, self).build(input_shape)

    def call(self, inputs, **kwargs):
        r"""Computes a loss from importance scores.

        Args:
            inputs: Importance tensor of shape ([batch], [N], K) .

        Returns:
            None.
        """
        # importances: ([batch], [N], K)
        importances = inputs

        loss = ops.mean(ops.abs(importances))
        loss = loss * self.factor
        self.add_loss(loss)
        return loss


from keras.layers import Layer, Dense, Activation, Add, Multiply
from kgcnn.layers.aggr import AggregateWeightedLocalEdges, AggregateLocalEdges
from kgcnn.layers.gather import GatherNodesOutgoing
from keras import ops
import kgcnn.ops.activ


class GCN(Layer):
    r"""Graph convolution according to `Kipf et al <https://arxiv.org/abs/1609.02907>`__ .

    Computes graph convolution as :math:`\sigma(A_s(WX+b))` where :math:`A_s` is the precomputed and scaled adjacency
    matrix. The scaled adjacency matrix is defined by :math:`A_s = D^{-0.5} (A + I) D^{-0.5}` with the degree
    matrix :math:`D` . In place of :math:`A_s` , this layers uses edge features (that are the entries of :math:`A_s` )
    and edge indices.

    .. note::

        :math:`A_s` is considered pre-scaled, this is not done by this layer!
        If no scaled edge features are available, you could consider use e.g. "mean",
        or :obj:`normalize_by_weights` to obtain a similar behaviour that is expected b
        y a pre-scaled adjacency matrix input.

    Edge features must be possible to broadcast to node features, since they are multiplied with the node features.
    Ideally they are weights of shape `(..., 1)` for broadcasting, e.g. entries of :math:`A_s` .
    """

    def __init__(self,
                 units,
                 pooling_method='scatter_sum',
                 normalize_by_weights=False,
                 activation="kgcnn>leaky_relu2",
                 use_bias=True,
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 kernel_initializer='glorot_uniform',
                 bias_initializer='zeros',
                 **kwargs):
        """Initialize layer.

        Args:
            units (int): Output dimension/ units of dense layer.
            pooling_method (str): Pooling method for summing edges. Default is 'segment_sum'.
            normalize_by_weights (bool): Normalize the pooled output by the sum of weights. Default is False.
                In this case the edge features are considered weights of dimension (...,1) and are summed for each node.
            activation (str): Activation. Default is "kgcnn>leaky_relu2".
            use_bias (bool): Use bias. Default is True.
            kernel_regularizer: Kernel regularization. Default is None.
            bias_regularizer: Bias regularization. Default is None.
            activity_regularizer: Activity regularization. Default is None.
            kernel_constraint: Kernel constrains. Default is None.
            bias_constraint: Bias constrains. Default is None.
            kernel_initializer: Initializer for kernels. Default is 'glorot_uniform'.
            bias_initializer: Initializer for bias. Default is 'zeros'.
        """
        super(GCN, self).__init__(**kwargs)
        # Changes in keras serialization behaviour for activations in 3.0.2.
        # Keep string at least for default. Also renames to prevent clashes with keras leaky_relu.
        if activation in ["kgcnn>leaky_relu", "kgcnn>leaky_relu2"]:
            activation = {"class_name": "function", "config": "kgcnn>leaky_relu2"}
        self.normalize_by_weights = normalize_by_weights
        self.pooling_method = pooling_method
        self.units = units
        kernel_args = {"kernel_regularizer": kernel_regularizer, "activity_regularizer": activity_regularizer,
                       "bias_regularizer": bias_regularizer, "kernel_constraint": kernel_constraint,
                       "bias_constraint": bias_constraint, "kernel_initializer": kernel_initializer,
                       "bias_initializer": bias_initializer, "use_bias": use_bias}
        pool_args = {"pooling_method": pooling_method, "normalize_by_weights": normalize_by_weights}

        # Layers
        self.layer_gather = GatherNodesOutgoing()
        self.layer_dense = Dense(units=self.units, activation='linear', **kernel_args)
        self.layer_pool = AggregateWeightedLocalEdges(**pool_args)
        self.layer_act = Activation(activation)
        
    def build(self, input_shape):
        super(GCN, self).build(input_shape)

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs: [nodes, edges, edge_index]

                - nodes (Tensor): Node embeddings of shape `(None, F)`
                - edges (Tensor): Edge or message embeddings of shape `(None, F)`
                - edge_index (Tensor): Edge indices referring to nodes of shape `(2, None)`

        Returns:
            Tensor: Node embeddings of shape `(None, F)`
        """
        node, edges, edge_index = inputs
        no = self.layer_dense(node, **kwargs)
        no = self.layer_gather([no, edge_index], **kwargs)
        nu = self.layer_pool([node, no, edge_index, edges], **kwargs)  # Summing for each node connection
        out = self.layer_act(nu, **kwargs)
        return out

    def get_config(self):
        """Update config."""
        config = super(GCN, self).get_config()
        config.update({"normalize_by_weights": self.normalize_by_weights,
                       "pooling_method": self.pooling_method, "units": self.units})
        conf_dense = self.layer_dense.get_config()
        for x in ["kernel_regularizer", "activity_regularizer", "bias_regularizer", "kernel_constraint",
                  "bias_constraint", "kernel_initializer", "bias_initializer", "use_bias"]:
            if x in conf_dense:
                config.update({x: conf_dense[x]})
        conf_act = self.layer_act.get_config()
        config.update({"activation": conf_act["activation"]})
        return config


class SchNetCFconv(Layer):
    r"""Continuous filter convolution of `SchNet <https://aip.scitation.org/doi/pdf/10.1063/1.5019779>`__ .

    Edges are processed by 2 :obj:`Dense` layers, multiplied on outgoing node features and pooled for receiving node.
    """

    def __init__(self, units,
                 cfconv_pool="scatter_sum",
                 use_bias=True,
                 activation="kgcnn>shifted_softplus",
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 kernel_initializer='glorot_uniform',
                 bias_initializer='zeros',
                 **kwargs):
        """Initialize Layer.

        Args:
            units (int): Units for Dense layer.
            cfconv_pool (str): Pooling method. Default is 'segment_sum'.
            use_bias (bool): Use bias. Default is True.
            activation (str): Activation function. Default is "kgcnn>shifted_softplus".
            kernel_regularizer: Kernel regularization. Default is None.
            bias_regularizer: Bias regularization. Default is None.
            activity_regularizer: Activity regularization. Default is None.
            kernel_constraint: Kernel constrains. Default is None.
            bias_constraint: Bias constrains. Default is None.
            kernel_initializer: Initializer for kernels. Default is 'glorot_uniform'.
            bias_initializer: Initializer for bias. Default is 'zeros'.
        """
        super(SchNetCFconv, self).__init__(**kwargs)
        # Changes in keras serialization behaviour for activations in 3.0.2.
        # Keep string at least for default.
        if activation in ["kgcnn>shifted_softplus"]:
            activation = {"class_name": "function", "config": "kgcnn>shifted_softplus"}
        self.cfconv_pool = cfconv_pool
        self.units = units
        self.use_bias = use_bias
        kernel_args = {"kernel_regularizer": kernel_regularizer, "activity_regularizer": activity_regularizer,
                       "bias_regularizer": bias_regularizer, "kernel_constraint": kernel_constraint,
                       "bias_constraint": bias_constraint, "kernel_initializer": kernel_initializer,
                       "bias_initializer": bias_initializer}
        # Layer
        self.lay_dense1 = Dense(units=self.units, activation=activation, use_bias=self.use_bias, **kernel_args)
        self.lay_dense2 = Dense(units=self.units, activation='linear', use_bias=self.use_bias, **kernel_args)
        self.lay_sum = AggregateLocalEdges(pooling_method=cfconv_pool)
        self.gather_n = GatherNodesOutgoing()
        self.lay_mult = Multiply()

    def build(self, input_shape):
        super(SchNetCFconv, self).build(input_shape)

    def call(self, inputs, **kwargs):
        """Forward pass. Calculate edge update.

        Args:
            inputs: [nodes, edges, edge_index]

                - nodes (Tensor): Node embeddings of shape ([N], F)
                - edges (Tensor): Edge or message embeddings of shape ([M], F)
                - edge_index (Tensor): Edge indices referring to nodes of shape (2, [M])

        Returns:
            Tensor: Updated node features.
        """
        node, edge, disjoint_indices = inputs
        x = self.lay_dense1(edge, **kwargs)
        x = self.lay_dense2(x, **kwargs)
        node2exp = self.gather_n([node, disjoint_indices], **kwargs)
        x = self.lay_mult([node2exp, x], **kwargs)
        x = self.lay_sum([node, x, disjoint_indices], **kwargs)
        return x

    def get_config(self):
        """Update layer config."""
        config = super(SchNetCFconv, self).get_config()
        config.update({"cfconv_pool": self.cfconv_pool, "units": self.units})
        config_dense = self.lay_dense1.get_config()
        for x in ["kernel_regularizer", "activity_regularizer", "bias_regularizer", "kernel_constraint",
                  "bias_constraint", "kernel_initializer", "bias_initializer", "activation", "use_bias"]:
            if x in config_dense:
                config.update({x: config_dense[x]})
        return config


class SchNetInteraction(Layer):
    r"""`SchNet <https://aip.scitation.org/doi/pdf/10.1063/1.5019779>`__ interaction block,
    which uses the continuous filter convolution from :obj:`SchNetCFconv` .
    """

    def __init__(self,
                 units=128,
                 cfconv_pool='scatter_sum',
                 use_bias=True,
                 activation="kgcnn>shifted_softplus",
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 kernel_initializer='glorot_uniform',
                 bias_initializer='zeros',
                 **kwargs):
        """Initialize Layer.

        Args:
            units (int): Dimension of node embedding. Default is 128.
            cfconv_pool (str): Pooling method information for SchNetCFconv layer. Default is 'scatter_sum'.
            use_bias (bool): Use bias in last layers. Default is True.
            activation (str): Activation function. Default is "kgcnn>shifted_softplus".
            kernel_regularizer: Kernel regularization. Default is None.
            bias_regularizer: Bias regularization. Default is None.
            activity_regularizer: Activity regularization. Default is None.
            kernel_constraint: Kernel constrains. Default is None.
            bias_constraint: Bias constrains. Default is None.
            kernel_initializer: Initializer for kernels. Default is 'glorot_uniform'.
            bias_initializer: Initializer for bias. Default is 'zeros'.
        """
        super(SchNetInteraction, self).__init__(**kwargs)
        # Changes in keras serialization behaviour for activations in 3.0.2.
        # Keep string at least for default.
        if activation in ["kgcnn>shifted_softplus"]:
            activation = {"class_name": "function", "config": "kgcnn>shifted_softplus"}
        self.cfconv_pool = cfconv_pool
        self.use_bias = use_bias
        self.units = units
        kernel_args = {"kernel_regularizer": kernel_regularizer, "activity_regularizer": activity_regularizer,
                       "bias_regularizer": bias_regularizer, "kernel_constraint": kernel_constraint,
                       "bias_constraint": bias_constraint, "kernel_initializer": kernel_initializer,
                       "bias_initializer": bias_initializer}
        conv_args = {"units": self.units, "use_bias": use_bias, "activation": activation, "cfconv_pool": cfconv_pool}

        # Layers
        self.lay_cfconv = SchNetCFconv(**conv_args, **kernel_args)
        self.lay_dense1 = Dense(units=self.units, activation='linear', use_bias=False, **kernel_args)
        self.lay_dense2 = Dense(units=self.units, activation=activation, use_bias=self.use_bias, **kernel_args)
        self.lay_dense3 = Dense(units=self.units, activation='linear', use_bias=self.use_bias, **kernel_args)
        self.lay_add = Add()

    def build(self, input_shape):
        super(SchNetInteraction, self).build(input_shape)

    def call(self, inputs, **kwargs):
        """Forward pass. Calculate node update.

        Args:
            inputs: [nodes, edges, tensor_index]

                - nodes (Tensor): Node embeddings of shape ([N], F)
                - edges (Tensor): Edge or message embeddings of shape ([M], F)
                - tensor_index (Tensor): Edge indices referring to nodes of shape (2, [M])

        Returns:
            Tensor: Updated node embeddings of shape ([N], F).
        """
        node, edge, indexlist = inputs
        x = self.lay_dense1(node, **kwargs)
        x = self.lay_cfconv([x, edge, indexlist], **kwargs)
        x = self.lay_dense2(x, **kwargs)
        x = self.lay_dense3(x, **kwargs)
        out = self.lay_add([node, x], **kwargs)
        return out

    def get_config(self):
        config = super(SchNetInteraction, self).get_config()
        config.update({"cfconv_pool": self.cfconv_pool, "units": self.units, "use_bias": self.use_bias})
        conf_dense = self.lay_dense2.get_config()
        for x in ["activation", "kernel_regularizer", "bias_regularizer", "activity_regularizer",
                  "kernel_constraint", "bias_constraint", "kernel_initializer", "bias_initializer"]:
            if x in conf_dense:
                config.update({x: conf_dense[x]})
        return config


class GIN(Layer):
    r"""Convolutional unit of `Graph Isomorphism Network from: How Powerful are Graph Neural Networks?
    <https://arxiv.org/abs/1810.00826>`__ .

    Computes graph convolution at step :math:`k` for node embeddings :math:`h_\nu` as:

    .. math::

        h_\nu^{(k)} = \phi^{(k)} ((1+\epsilon^{(k)}) h_\nu^{k-1} + \sum_{u\in N(\nu)}) h_u^{k-1}.

    with optional learnable :math:`\epsilon^{(k)}`

    .. note::

        The non-linear mapping :math:`\phi^{(k)}` , usually an :obj:`MLP` , is not included in this layer.
    """

    def __init__(self,
                 pooling_method='scatter_sum',
                 epsilon_learnable=False,
                 **kwargs):
        """Initialize layer.

        Args:
            epsilon_learnable (bool): If epsilon is learnable or just constant zero. Default is False.
            pooling_method (str): Pooling method for summing edges. Default is 'segment_sum'.
        """
        super(GIN, self).__init__(**kwargs)
        self.pooling_method = pooling_method
        self.epsilon_learnable = epsilon_learnable

        # Layers
        self.lay_gather = GatherNodesOutgoing()
        self.lay_pool = AggregateLocalEdges(pooling_method=self.pooling_method)
        self.lay_add = Add()

        # Epsilon with trainable as optional and default zeros initialized.
        self.eps_k = self.add_weight(shape=tuple(), name="epsilon_k", trainable=self.epsilon_learnable,
                                     initializer="zeros", dtype=self.dtype)

    def build(self, input_shape):
        """Build layer."""
        super(GIN, self).build(input_shape)

    def call(self, inputs, **kwargs):
        r"""Forward pass.

        Args:
            inputs: [nodes, edge_index]

                - nodes (Tensor): Node embeddings of shape `([N], F)`
                - edge_index (Tensor): Edge indices referring to nodes of shape `(2, [M])`

        Returns:
            Tensor: Node embeddings of shape `([N], F)`
        """
        node, edge_index = inputs
        ed = self.lay_gather([node, edge_index], **kwargs)
        nu = self.lay_pool([node, ed, edge_index], **kwargs)  # Summing for each node connection
        no = (ops.convert_to_tensor(1, dtype=self.eps_k.dtype) + self.eps_k) * node
        out = self.lay_add([no, nu], **kwargs)
        return out

    def get_config(self):
        """Update config."""
        config = super(GIN, self).get_config()
        config.update({"pooling_method": self.pooling_method,
                       "epsilon_learnable": self.epsilon_learnable})
        return config


class GINE(Layer):
    r"""Convolutional unit of `Strategies for Pre-training Graph Neural Networks <https://arxiv.org/abs/1905.12265>`__ .

    Computes graph convolution with node embeddings :math:`\mathbf{h}` and compared to :obj:`GIN_conv`,
    adds edge embeddings of :math:`\mathbf{e}_{ij}`.

    .. math::

        \mathbf{h}^{\prime}_i = f_{\mathbf{\Theta}} \left( (1 + \epsilon) \cdot
        \mathbf{h}_i + \sum_{j \in \mathcal{N}(i)} \phi \; ( \mathbf{h}_j + \mathbf{e}_{ij} ) \right),

    with optionally learnable :math:`\epsilon`. The activation :math:`\phi` can be chosen differently
    but defaults to RELU.

    .. note::

        The final non-linear mapping :math:`f_{\mathbf{\Theta}}`, usually an :obj:`MLP`, is not included in this layer.
    """

    def __init__(self,
                 pooling_method='scatter_sum',
                 epsilon_learnable=False,
                 activation="relu",
                 activity_regularizer=None,
                 **kwargs):
        """Initialize layer.

        Args:
            epsilon_learnable (bool): If epsilon is learnable or just constant zero. Default is False.
            pooling_method (str): Pooling method for summing edges. Default is 'segment_sum'.
            activation: Activation function, such as `tf.nn.relu`, or string name of
                built-in activation function, such as "relu".
            activity_regularizer: Regularizer function applied to
                the output of the layer (its "activation"). Default is None.
        """
        super(GINE, self).__init__(**kwargs)
        self.pooling_method = pooling_method
        self.epsilon_learnable = epsilon_learnable

        # Layers
        self.layer_gather = GatherNodesOutgoing()
        self.layer_pool = AggregateLocalEdges(pooling_method=self.pooling_method)
        self.layer_add = Add()
        self.layer_act = Activation(activation=activation,
                                    activity_regularizer=activity_regularizer)

        # Epsilon with trainable as optional and default zeros initialized.
        self.eps_k = self.add_weight(shape=tuple(), name="epsilon_k", trainable=self.epsilon_learnable,
                                     initializer="zeros", dtype=self.dtype)

    def build(self, input_shape):
        """Build layer."""
        super(GINE, self).build(input_shape)

    def call(self, inputs, **kwargs):
        r"""Forward pass.

        Args:
            inputs: [nodes, edge_index, edges]

                - nodes (Tensor): Node embeddings of shape `([N], F)`
                - edge_index (Tensor): Edge indices referring to nodes of shape `(2, [M])`
                - edges (Tensor): Edge embeddings for index tensor of shape `([M], F)`

        Returns:
            Tensor: Node embeddings of shape `([N], F)`
        """
        node, edge_index, edges = inputs
        ed = self.layer_gather([node, edge_index], **kwargs)
        ed = self.layer_add([ed, edges])
        ed = self.layer_act(ed)
        nu = self.layer_pool([node, ed, edge_index], **kwargs)  # Summing for each node connection
        no = (ops.convert_to_tensor(1, dtype=self.eps_k.dtype) + self.eps_k)*node
        out = self.layer_add([no, nu], **kwargs)
        return out

    def get_config(self):
        """Update config."""
        config = super(GINE, self).get_config()
        config.update({"pooling_method": self.pooling_method,
                       "epsilon_learnable": self.epsilon_learnable})
        conf_act = self.layer_act.get_config()
        for x in ["activation", "activity_regularizer"]:
            if x in conf_act:
                config.update({x: conf_act[x]})
        return config



import os.path
import numpy as np
import tensorflow as tf
import h5py
from typing import List, Union


def _check_for_inner_shape(array_list: List[np.ndarray]) -> Union[None, tuple, list]:
    """Simple function to verify inner shape for list of numpy arrays."""
    # Cannot find inner shape for empty list.
    if len(array_list) == 0:
        return None
    # For fast check all items must be numpy arrays to get the inner shape easily.
    if not all(isinstance(x, np.ndarray) for x in array_list):
        return None
    shapes = [x.shape for x in array_list]
    # Must have all same rank.
    if not all(len(x) == len(shapes[0]) for x in shapes):
        return None
    # All Empty. No inner shape.
    if len(shapes[0]) == 0:
        return None
    # Empty inner shape.
    if len(shapes[0]) <= 1:
        return tuple([])
    # If all same inner shape.
    if all(x[1:] == shapes[0][1:] for x in shapes):
        return shapes[0][1:]


class RaggedTensorNumpyFile:
    """Class representing a NumPy '.npz' file to store a ragged tensor on disk.

    For the moment only ragged tensors of ragged rank of one are supported. However, arbitrary ragged tensors can be
    supported in principle.
    """

    _device = '/cpu:0'

    def __init__(self, file_path: str, compressed: bool = False):
        """Make class for a NPZ file.

        Args:
            file_path (str): Path to file on disk.
            compressed (bool): Whether to use compression.
        """
        self.file_path = file_path
        self.compressed = compressed

    def write(self, ragged_array: Union[tf.RaggedTensor, List[np.ndarray], list]):
        """Write ragged array to file.

        .. code-block:: python

            from kgcnn.io.file import RaggedTensorNumpyFile
            import numpy as np
            data = [np.array([[0, 1],[0, 2]]), np.array([[1, 1]]), np.array([[0, 1],[2, 2], [0, 3]])]
            f = RaggedTensorNumpyFile("test.npz")
            f.write(data)
            print(f.read())

        Args:
            ragged_array (list, tf.RaggedTensor): List or list of numpy arrays.

        Returns:
            None.
        """
        # We use tensorflow functions to ensure an eager ragged tensor.
        if not isinstance(ragged_array, tf.RaggedTensor):
            with tf.device(self._device):
                ragged_array = tf.ragged.constant(ragged_array, inner_shape=_check_for_inner_shape(ragged_array))
        assert ragged_array.ragged_rank == 1, "Only support for ragged_rank=1 at the moment."
        values = np.array(ragged_array.values)
        row_splits = np.array(ragged_array.row_splits)
        # Since the shape array can not have nones, we convert nones to 0.
        # Not ideal, but could make an extra shape array to indicate ragged dimensions.
        shape = np.array([x if x is not None else 0 for x in ragged_array.shape], dtype="uint64")
        ragged_rank = np.array(ragged_array.ragged_rank)
        rank = np.array(len(shape))
        out = {"values": values,
               "row_splits": row_splits,
               "shape": shape,
               "ragged_rank": ragged_rank,
               "rank": rank}
        if self.compressed:
            np.savez_compressed(self.file_path, **out)
        else:
            np.savez(self.file_path, **out)

    def read(self, return_as_tensor: bool = False):
        """Read the file into memory.

        Args:
            return_as_tensor: Whether to return tf.RaggedTensor.

        Returns:
            tf.RaggedTensor: Ragged tensor form file.
        """
        # Here only ragged rank one loading is supported.
        data = np.load(self.file_path)
        values = data.get("values")
        row_splits = data.get("row_splits")
        if return_as_tensor:
            with tf.device(self._device):
                out = tf.RaggedTensor.from_row_splits(values, row_splits)
            return out
        return np.split(values, row_splits[1:-1])

    def __getitem__(self, item):
        """Get single item from the ragged tensor on file.

        Args:
            item (int): Index of the item to get.
        """
        assert isinstance(item, int), "Only single index is supported, no slicing."
        # NOTE: At the moment mmap is not supported for NPZ files.
        with np.load(self.file_path, mmap_mode="r") as data:
            row_splits = np.array(data.get("row_splits"))
            out_data = np.array(data["values"][row_splits[item]:row_splits[item + 1]])
        return out_data

    def exists(self):
        """Check if file for path information of this class exists."""
        return os.path.exists(self.file_path)

    def __len__(self):
        """Length of the tensor on file."""
        data = np.load(self.file_path)
        row_splits = data.get("row_splits")
        num_row_splits = int(row_splits.shape[0])
        # length is num_row_splits - 1
        return num_row_splits-1


class RaggedTensorHDFile:
    """Class representing an HDF '.hdf5' file to store a ragged tensor on disk.

    For the moment only ragged tensors of ragged rank of one are supported. However, arbitrary ragged tensors can be
    supported in principle.
    """

    _device = '/cpu:0'

    def __init__(self, file_path: str, compressed: bool = None):
        """Make class for a HDF5 file.

        Args:
            file_path (str): Path to file on disk.
            compressed: Compression to use. Not used at the moment.
        """
        self.file_path = file_path
        self.compressed = compressed

    def write(self, ragged_array: List[np.ndarray]):
        """Write ragged array to file.

        .. code-block:: python

            from kgcnn.io.file import RaggedTensorHDFile
            import numpy as np
            data = [np.array([[0, 1],[0, 2]]), np.array([[1, 1]]), np.array([[0, 1],[2, 2], [0, 3]])]
            f = RaggedTensorHDFile("test.hdf5")
            f.write(data)
            print(f.read())

        Args:
            ragged_array (list, tf.RaggedTensor): List or list of numpy arrays.

        Returns:
            None.
        """
        # We use tensorflow functions to ensure an eager ragged tensor.
        if not isinstance(ragged_array, tf.RaggedTensor):
            with tf.device(self._device):
                ragged_array = tf.ragged.constant(ragged_array, inner_shape=_check_for_inner_shape(ragged_array))
        assert ragged_array.ragged_rank == 1, "Only support for ragged_rank=1 at the moment."
        values = np.array(ragged_array.values)
        row_splits = np.array(ragged_array.row_splits)
        # Since the shape array can not have nones, we convert nones to 0.
        # Not ideal, but could make an extra shape array to indicate ragged dimensions.
        shape = np.array([x if x is not None else 0 for x in ragged_array.shape], dtype="uint64")
        ragged_rank = np.array(ragged_array.ragged_rank)
        rank = np.array(len(shape))
        with h5py.File(self.file_path, "w") as file:
            file.create_dataset("values", data=values,
                                maxshape=[x if i > 0 else None for i, x in enumerate(values.shape)])
            file.create_dataset("row_splits", data=row_splits, maxshape=(None, ))
            file.create_dataset("shape", data=shape)
            file.create_dataset("rank", data=rank)
            file.create_dataset("ragged_rank", data=ragged_rank)

    def read(self, return_as_tensor: bool = False):
        """Read the file into memory.

        Args:
            return_as_tensor: Whether to return tf.RaggedTensor.

        Returns:
            tf.RaggedTensor: Ragged tensor form file.
        """
        with h5py.File(self.file_path, "r") as file:
            values = file["values"]
            row_splits = file["row_splits"]
            if return_as_tensor:
                with tf.device(self._device):
                    out = tf.RaggedTensor.from_row_splits(np.array(values), np.array(row_splits))
            else:
                out = np.split(values, row_splits[1:-1])
        return out

    def __getitem__(self, item: int):
        """Get single item from the ragged tensor on file.

        Args:
            item (int): Index of the item to get.
        """
        assert isinstance(item, int), "Only single index is supported, no slicing."
        with h5py.File(self.file_path, "r") as file:
            row_splits = file["row_splits"]
            out_data = np.array(file["values"][row_splits[item]:row_splits[item+1]])
        return out_data

    def append(self, item):
        """Append single item to ragged tensor.

        Args:
            item (np.ndarray, tf.Tensor): Item to append.

        Returns:
            None.
        """
        with h5py.File(self.file_path, "r+") as file:
            file["values"].resize(
                file["values"].shape[0] + len(item), axis=0
            )
            split_last = file["row_splits"][-1]
            file["row_splits"].resize(
                file["row_splits"].shape[0] + 1, axis=0
            )
            len_last = len(item)
            file["row_splits"][-1] = split_last + len_last
            file["values"][split_last:split_last+len_last] = item

    def append_multiple(self, items: list):
        """Append multiple items to ragged tensor.

        Args:
            items (list): List of items to append. Must match in shape.

        Returns:
            None.
        """
        new_values = np.concatenate(items, axis=0)
        new_len = len(items)
        new_splits = np.cumsum([len(x) for x in items])
        with h5py.File(self.file_path, "r+") as file:
            file["values"].resize(
                file["values"].shape[0] + new_values.shape[0], axis=0
            )
            split_last = file["row_splits"][-1]
            file["row_splits"].resize(
                file["row_splits"].shape[0] + new_len, axis=0
            )
            file["row_splits"][-new_len:] = split_last + new_splits
            file["values"][split_last:+split_last+new_splits[-1]] = new_values

    def __len__(self):
        """Length of the tensor on file."""
        with h5py.File(self.file_path, "r") as file:
            num_row_splits = int(file["row_splits"].shape[0])
        # length is num_row_splits - 1
        return num_row_splits-1

    def exists(self):
        """Check if file for path information of this class exists."""
        return os.path.exists(self.file_path)



import keras as ks
from kgcnn.layers.scale import get as get_scaler
from ._model import model_disjoint, model_disjoint_crystal
from kgcnn.layers.modules import Input
from kgcnn.models.casting import (template_cast_output, template_cast_list_input,
                                  template_cast_list_input_docs, template_cast_output_docs)
from kgcnn.models.utils import update_model_kwargs
from keras.backend import backend as backend_to_use

# To be updated if model is changed in a significant way.
__model_version__ = "2023-12-04"

# Supported backends
__kgcnn_model_backend_supported__ = ["tensorflow", "torch", "jax"]
if backend_to_use() not in __kgcnn_model_backend_supported__:
    raise NotImplementedError("Backend '%s' for model 'DimeNetPP' is not supported." % backend_to_use())

# Implementation of DimeNet++ in `keras` from paper:
# Fast and Uncertainty-Aware Directional Message Passing for Non-Equilibrium Molecules
# Johannes Klicpera, Shankari Giri, Johannes T. Margraf, Stephan Günnemann
# https://arxiv.org/abs/2011.14115
# Original code: https://github.com/gasteigerjo/dimenet

model_default = {
    "name": "DimeNetPP",
    "inputs": [
        {"shape": [None], "name": "node_number", "dtype": "int64"},
        {"shape": [None, 3], "name": "node_coordinates", "dtype": "float32"},
        {"shape": [None, 2], "name": "edge_indices", "dtype": "int64"},
        {"shape": [None, 2], "name": "angle_indices", "dtype": "int64"},
        {"shape": (), "name": "total_nodes", "dtype": "int64"},
        {"shape": (), "name": "total_edges", "dtype": "int64"},
        {"shape": (), "name": "total_angles", "dtype": "int64"}
    ],
    "input_tensor_type": "padded",
    "input_embedding": None,  # deprecated
    "cast_disjoint_kwargs": {},
    "input_node_embedding": {
        "input_dim": 95, "output_dim": 128, "embeddings_initializer": {
            "class_name": "RandomUniform",
            "config": {"minval": -1.7320508075688772, "maxval": 1.7320508075688772}}
    },
    "emb_size": 128, "out_emb_size": 256, "int_emb_size": 64, "basis_emb_size": 8,
    "num_blocks": 4, "num_spherical": 7, "num_radial": 6,
    "cutoff": 5.0, "envelope_exponent": 5,
    "num_before_skip": 1, "num_after_skip": 2, "num_dense_output": 3,
    "num_targets": 64, "extensive": True, "output_init": "zeros",
    "activation": "swish", "verbose": 10,
    "output_embedding": "graph",
    "use_output_mlp": True,
    "output_tensor_type": "padded",
    "output_scaling": None,
    "output_mlp": {"use_bias": [True, False],
                   "units": [64, 12], "activation": ["swish", "linear"]}
}


@update_model_kwargs(model_default, update_recursive=0, deprecated=["input_embedding", "output_to_tensor"])
def make_model(inputs: list = None,
               input_tensor_type: str = None,
               cast_disjoint_kwargs: dict = None,
               input_embedding: dict = None,
               input_node_embedding: dict = None,
               emb_size: int = None,
               out_emb_size: int = None,
               int_emb_size: int = None,
               basis_emb_size: int = None,
               num_blocks: int = None,
               num_spherical: int = None,
               num_radial: int = None,
               cutoff: float = None,
               envelope_exponent: int = None,
               num_before_skip: int = None,
               num_after_skip: int = None,
               num_dense_output: int = None,
               num_targets: int = None,
               activation: str = None,
               extensive: bool = None,
               output_init: str = None,
               verbose: int = None,
               name: str = None,
               output_embedding: str = None,
               output_tensor_type: str = None,
               use_output_mlp: bool = None,
               output_mlp: dict = None,
               output_scaling: dict = None
               ):
    """Make `DimeNetPP <https://arxiv.org/abs/2011.14115>`_ graph network via functional API.
    Default parameters can be found in :obj:`kgcnn.literature.DimeNetPP.model_default`.

    **Model inputs**:
    Model uses the list template of inputs and standard output template.
    The supported inputs are  :obj:`[nodes, coordinates, edge_indices, angle_indices...]`
    with '...' indicating mask or ID tensors following the template below.
    Note that you must supply angle indices as index pairs that refer to two edges.

    %s

    **Model outputs**:
    The standard output template:

    %s


    Args:
        inputs (list): List of dictionaries unpacked in :obj:`tf.keras.layers.Input`. Order must match model definition.
        input_tensor_type (str): Input type of graph tensor. Default is "padded".
        cast_disjoint_kwargs (dict): Dictionary of arguments for casting layer.
        input_embedding (dict): Deprecated in favour of input_node_embedding etc.
        input_node_embedding (dict): Dictionary of embedding arguments for nodes unpacked in :obj:`Embedding` layers.
        emb_size (int): Overall embedding size used for the messages.
        out_emb_size (int): Embedding size for output of :obj:`DimNetOutputBlock`.
        int_emb_size (int): Embedding size used for interaction triplets.
        basis_emb_size (int): Embedding size used inside the basis transformation.
        num_blocks (int): Number of graph embedding blocks or depth of the network.
        num_spherical (int): Number of spherical components in :obj:`SphericalBasisLayer`.
        num_radial (int): Number of radial components in basis layer.
        cutoff (float): Distance cutoff for basis layer.
        envelope_exponent (int): Exponent in envelope function for basis layer.
        num_before_skip (int): Number of residual layers in interaction block before skip connection
        num_after_skip (int): Number of residual layers in interaction block after skip connection
        num_dense_output (int): Number of dense units in output :obj:`DimNetOutputBlock`.
        num_targets (int): Number of targets or output embedding dimension of the model.
        activation (str, dict): Activation to use.
        extensive (bool): Graph output for extensive target to apply sum for pooling or mean otherwise.
        output_init (str, dict): Output initializer for kernel.
        verbose (int): Level of verbosity.
        name (str): Name of the model.
        output_embedding (str): Main embedding task for graph network. Either "node", "edge" or "graph".
        use_output_mlp (bool): Whether to use the final output MLP. Possibility to skip final :obj:`MLP`.
        output_mlp (dict): Dictionary of layer arguments unpacked in the final classification :obj:`MLP` layer block.
            Defines number of model outputs and activation. Note that DimeNetPP originally defines the output dimension
            via `num_targets`. But this can be set to `out_emb_size` and the `output_mlp` be used for more
            specific control.
        output_scaling (dict): Dictionary of layer arguments unpacked in scaling layers. Default is None.
        output_tensor_type (str): Output type of graph tensors such as nodes or edges. Default is "padded".

    Returns:
        :obj:`keras.models.Model`
    """
    # Make input
    model_inputs = [Input(**x) for x in inputs]

    dj = template_cast_list_input(
        model_inputs,
        input_tensor_type=input_tensor_type,
        cast_disjoint_kwargs=cast_disjoint_kwargs,
        mask_assignment=[0, 0, 1, 2],
        index_assignment=[None, None, 0, 2]
    )

    n, x, edi, adi, batch_id_node, batch_id_edge, batch_id_angles, node_id, edge_id, angle_id, count_nodes, count_edges, count_angles = dj

    out = model_disjoint(
        [n, x, edi, adi, batch_id_node, count_nodes],
        use_node_embedding=("int" in inputs[0]['dtype']) if input_node_embedding is not None else False,
        input_node_embedding=input_node_embedding,
        emb_size=emb_size,
        out_emb_size=out_emb_size,
        int_emb_size=int_emb_size,
        basis_emb_size=basis_emb_size,
        num_blocks=num_blocks,
        num_spherical=num_spherical,
        num_radial=num_radial,
        cutoff=cutoff,
        envelope_exponent=envelope_exponent,
        num_before_skip=num_before_skip,
        num_after_skip=num_after_skip,
        num_dense_output=num_dense_output,
        num_targets=num_targets,
        activation=activation,
        extensive=extensive,
        output_init=output_init,
        use_output_mlp=use_output_mlp,
        output_embedding=output_embedding,
        output_mlp=output_mlp
    )

    if output_scaling is not None:
        scaler = get_scaler(output_scaling["name"])(**output_scaling)
        if scaler.extensive:
            # Node information must be numbers, or we need an additional input.
            out = scaler([out, n, batch_id_node])
        else:
            out = scaler(out)

    # Output embedding choice
    out = template_cast_output(
        [out, batch_id_node, batch_id_edge, node_id, edge_id, count_nodes, count_edges],
        output_embedding=output_embedding, output_tensor_type=output_tensor_type,
        input_tensor_type=input_tensor_type, cast_disjoint_kwargs=cast_disjoint_kwargs,
    )

    model = ks.models.Model(inputs=model_inputs, outputs=out, name=name)

    model.__kgcnn_model_version__ = __model_version__

    if output_scaling is not None:
        def set_scale(*args, **kwargs):
            scaler.set_scale(*args, **kwargs)

        setattr(model, "set_scale", set_scale)

    return model


make_model.__doc__ = make_model.__doc__ % (template_cast_list_input_docs, template_cast_output_docs)

model_crystal_default = {
    "name": "DimeNetPP",
    "inputs": [
        {"shape": [None], "name": "node_number", "dtype": "int64", "ragged": True},
        {"shape": [None, 3], "name": "node_coordinates", "dtype": "float32", "ragged": True},
        {"shape": [None, 2], "name": "edge_indices", "dtype": "int64", "ragged": True},
        {"shape": [None, 2], "name": "angle_indices", "dtype": "int64", "ragged": True},
        {'shape': (None, 3), 'name': "edge_image", 'dtype': 'int64', 'ragged': True},
        {'shape': (3, 3), 'name': "graph_lattice", 'dtype': 'float32', 'ragged': False}
    ],
    "input_tensor_type": "ragged",
    "input_embedding": None,  # deprecated
    "cast_disjoint_kwargs": {},
    "input_node_embedding": {
        "input_dim": 95, "output_dim": 128, "embeddings_initializer": {
            "class_name": "RandomUniform",
            "config": {"minval": -1.7320508075688772, "maxval": 1.7320508075688772}}
    },
    "emb_size": 128, "out_emb_size": 256, "int_emb_size": 64, "basis_emb_size": 8,
    "num_blocks": 4, "num_spherical": 7, "num_radial": 6,
    "cutoff": 5.0, "envelope_exponent": 5,
    "num_before_skip": 1, "num_after_skip": 2, "num_dense_output": 3,
    "num_targets": 64, "extensive": True, "output_init": "zeros",
    "activation": "swish", "verbose": 10,
    "output_embedding": "graph",
    "use_output_mlp": True,
    "output_tensor_type": "padded",
    "output_scaling": None,
    "output_mlp": {"use_bias": [True, False],
                   "units": [64, 12], "activation": ["swish", "linear"]}
}


@update_model_kwargs(model_crystal_default, update_recursive=0, deprecated=["input_embedding", "output_to_tensor"])
def make_crystal_model(inputs: list = None,
                       input_tensor_type: str = None,
                       cast_disjoint_kwargs: dict = None,
                       input_embedding: dict = None,
                       input_node_embedding: dict = None,
                       emb_size: int = None,
                       out_emb_size: int = None,
                       int_emb_size: int = None,
                       basis_emb_size: int = None,
                       num_blocks: int = None,
                       num_spherical: int = None,
                       num_radial: int = None,
                       cutoff: float = None,
                       envelope_exponent: int = None,
                       num_before_skip: int = None,
                       num_after_skip: int = None,
                       num_dense_output: int = None,
                       num_targets: int = None,
                       activation: str = None,
                       extensive: bool = None,
                       output_init: str = None,
                       verbose: int = None,
                       name: str = None,
                       output_embedding: str = None,
                       output_tensor_type: str = None,
                       use_output_mlp: bool = None,
                       output_mlp: dict = None,
                       output_scaling: dict = None
                       ):
    """Make `DimeNetPP <https://arxiv.org/abs/2011.14115>`__ graph network via functional API.
    Default parameters can be found in :obj:`kgcnn.literature.DimeNetPP.model_crystal_default`.

    .. note::

        DimeNetPP does require a large amount of memory for this implementation, which increase quickly with
        the number of connections in a batch. Use ragged input or dataloader if possible.

    **Model inputs**:
    Model uses the list template of inputs and standard output template.
    The supported inputs are  :obj:`[nodes, coordinates, edge_indices, angle_indices, image_translation, lattice, ...]`
    with '...' indicating mask or ID tensors following the template below.
    Note that you must supply angle indices as index pairs that refer to two edges.

    %s

    **Model outputs**:
    The standard output template:

    %s


    Args:
        inputs (list): List of dictionaries unpacked in :obj:`tf.keras.layers.Input`. Order must match model definition.
        input_tensor_type (str): Input type of graph tensor. Default is "padded".
        cast_disjoint_kwargs (dict): Dictionary of arguments for casting layer.
        input_embedding (dict): Deprecated in favour of input_node_embedding etc.
        input_node_embedding (dict): Dictionary of embedding arguments for nodes unpacked in :obj:`Embedding` layers.
        emb_size (int): Overall embedding size used for the messages.
        out_emb_size (int): Embedding size for output of :obj:`DimNetOutputBlock`.
        int_emb_size (int): Embedding size used for interaction triplets.
        basis_emb_size (int): Embedding size used inside the basis transformation.
        num_blocks (int): Number of graph embedding blocks or depth of the network.
        num_spherical (int): Number of spherical components in :obj:`SphericalBasisLayer`.
        num_radial (int): Number of radial components in basis layer.
        cutoff (float): Distance cutoff for basis layer.
        envelope_exponent (int): Exponent in envelope function for basis layer.
        num_before_skip (int): Number of residual layers in interaction block before skip connection
        num_after_skip (int): Number of residual layers in interaction block after skip connection
        num_dense_output (int): Number of dense units in output :obj:`DimNetOutputBlock`.
        num_targets (int): Number of targets or output embedding dimension of the model.
        activation (str, dict): Activation to use.
        extensive (bool): Graph output for extensive target to apply sum for pooling or mean otherwise.
        output_init (str, dict): Output initializer for kernel.
        verbose (int): Level of verbosity.
        name (str): Name of the model.
        output_embedding (str): Main embedding task for graph network. Either "node", "edge" or "graph".
        use_output_mlp (bool): Whether to use the final output MLP. Possibility to skip final :obj:`MLP`.
        output_mlp (dict): Dictionary of layer arguments unpacked in the final classification :obj:`MLP` layer block.
            Defines number of model outputs and activation. Note that DimeNetPP originally defines the output dimension
            via `num_targets`. But this can be set to `out_emb_size` and the `output_mlp` be used for more
            specific control.
        output_scaling (dict): Dictionary of layer arguments unpacked in scaling layers. Default is None.
        output_tensor_type (str): Output type of graph tensors such as nodes or edges. Default is "padded".

    Returns:
        :obj:`keras.models.Model`
    """
    # Make input
    model_inputs = [Input(**x) for x in inputs]

    disjoint_inputs = template_cast_list_input(
        model_inputs, input_tensor_type=input_tensor_type,
        cast_disjoint_kwargs=cast_disjoint_kwargs,
        index_assignment=[None, None, 0, 2, None, None],
        mask_assignment=[0, 0, 1, 2, 1, None]
    )
    n, x, edi, angi, img, lattice, batch_id_node, batch_id_edge, batch_id_angles, node_id, edge_id, angle_id, count_nodes, count_edges, count_angles = disjoint_inputs

    # Wrapp disjoint model
    out = model_disjoint_crystal(
        [n, x, edi, angi, img, lattice, batch_id_node, batch_id_edge, count_nodes],
        use_node_embedding=("int" in inputs[0]['dtype']) if input_node_embedding is not None else False,
        input_node_embedding=input_node_embedding,
        emb_size=emb_size,
        out_emb_size=out_emb_size,
        int_emb_size=int_emb_size,
        basis_emb_size=basis_emb_size,
        num_blocks=num_blocks,
        num_spherical=num_spherical,
        num_radial=num_radial,
        cutoff=cutoff,
        envelope_exponent=envelope_exponent,
        num_before_skip=num_before_skip,
        num_after_skip=num_after_skip,
        num_dense_output=num_dense_output,
        num_targets=num_targets,
        activation=activation,
        extensive=extensive,
        output_init=output_init,
        use_output_mlp=use_output_mlp,
        output_embedding=output_embedding,
        output_mlp=output_mlp
    )

    if output_scaling is not None:
        scaler = get_scaler(output_scaling["name"])(**output_scaling)
        if scaler.extensive:
            # Node information must be numbers, or we need an additional input.
            out = scaler([out, n, batch_id_node])
        else:
            out = scaler(out)

    # Output embedding choice
    out = template_cast_output(
        [out, batch_id_node, batch_id_edge, node_id, edge_id, count_nodes, count_edges],
        output_embedding=output_embedding, output_tensor_type=output_tensor_type,
        input_tensor_type=input_tensor_type, cast_disjoint_kwargs=cast_disjoint_kwargs,
    )

    model = ks.models.Model(inputs=model_inputs, outputs=out, name=name)

    model.__kgcnn_model_version__ = __model_version__

    if output_scaling is not None:
        def set_scale(*args, **kwargs):
            scaler.set_scale(*args, **kwargs)

        setattr(model, "set_scale", set_scale)

    return model


make_crystal_model.__doc__ = make_crystal_model.__doc__ % (template_cast_list_input_docs, template_cast_output_docs)



import keras as ks
from kgcnn.layers.modules import Input
from kgcnn.models.utils import update_model_kwargs
from keras.backend import backend as backend_to_use
from kgcnn.layers.scale import get as get_scaler
from kgcnn.models.casting import (template_cast_output, template_cast_list_input,
                                  template_cast_list_input_docs, template_cast_output_docs)
from ._model import model_disjoint_crystal

# To be updated if model is changed in a significant way.
__model_version__ = "2023-11-28"

# Supported backends
__kgcnn_model_backend_supported__ = ["tensorflow", "torch", "jax"]

if backend_to_use() not in __kgcnn_model_backend_supported__:
    raise NotImplementedError("Backend '%s' for model 'PAiNN' is not supported." % backend_to_use())

# Implementation of CGCNN in `keras` from paper:
# Crystal Graph Convolutional Neural Networks for an Accurate and Interpretable Prediction of Material Properties
# Tian Xie and Jeffrey C. Grossman
# https://journals.aps.org/prl/abstract/10.1103/PhysRevLett.120.145301


model_crystal_default = {
    'name': 'CGCNN',
    'inputs': [
        {'shape': (None,), 'name': 'node_number', 'dtype': 'int64'},
        {'shape': (None, 3), 'name': 'node_frac_coordinates', 'dtype': 'float64'},
        {'shape': (None, 2), 'name': 'edge_indices', 'dtype': 'int64'},
        {'shape': (None, 3), 'name': 'cell_translations', 'dtype': 'float32'},
        {'shape': (3, 3), 'name': 'lattice_matrix', 'dtype': 'float64'},
        # {'shape': (None, 1), 'name': 'multiplicities', 'dtype': 'float32'},  # For asu"
        # {'shape': (None, 4, 4), 'name': 'symmops', 'dtype': 'float64'},
        {"shape": (), "name": "total_nodes", "dtype": "int64"},
        {"shape": (), "name": "total_edges", "dtype": "int64"}
    ],
    "input_tensor_type": "padded",
    "input_embedding": None,  # deprecated
    "cast_disjoint_kwargs": {},
    "input_node_embedding": {"input_dim": 95, "output_dim": 64},
    'representation': 'unit',  # None, 'asu' or 'unit'
    'expand_distance': True,
    'make_distances': True,
    'gauss_args': {'bins': 40, 'distance': 8, 'offset': 0.0, 'sigma': 0.4},
    'depth': 3,
    "verbose": 10,
    'conv_layer_args': {
        'units': 64,
        'activation_s': 'softplus',
        'activation_out': 'softplus',
        'batch_normalization': True,
    },
    'node_pooling_args': {'pooling_method': 'scatter_mean'},
    "output_embedding": "graph",
    "output_to_tensor": None,  # deprecated
    "output_scaling": None,
    "output_tensor_type": "padded",
    'output_mlp': {'use_bias': [True, False], 'units': [64, 1],
                   'activation': ['softplus', 'linear']},
}


@update_model_kwargs(model_crystal_default, update_recursive=0, deprecated=["input_embedding", "output_to_tensor"])
def make_crystal_model(inputs: list = None,
                       input_tensor_type: str = None,
                       input_embedding: dict = None,
                       input_node_embedding: dict = None,
                       cast_disjoint_kwargs: dict = None,
                       representation: str = None,
                       make_distances: bool = None,
                       conv_layer_args: dict = None,
                       expand_distance: bool = None,
                       depth: int = None,
                       name: str = None,
                       verbose: int = None,
                       gauss_args: dict = None,
                       node_pooling_args: dict = None,
                       output_to_tensor: dict = None,
                       output_mlp: dict = None,
                       output_embedding: str = None,
                       output_scaling: dict = None,
                       output_tensor_type: str = None
                       ):
    r"""Make `CGCNN <https://journals.aps.org/prl/abstract/10.1103/PhysRevLett.120.145301>`__ graph network
    via functional API.
    Default parameters can be found in :obj:`kgcnn.literature.CGCNN.model_crystal_default`.

    **Model inputs**:
    Model uses the list template of inputs and standard output template.
    Model supports :obj:`[node_attributes, node_frac_coordinates, bond_indices, lattice, cell_translations, ...]`
    if representation='unit'` and `make_distances=True` or
    :obj:`[node_attributes, node_frac_coords, bond_indices, lattice, cell_translations, multiplicities, symmops, ...]`
    if `representation='asu'` and `make_distances=True`
    or :obj:`[node_attributes, edge_distance, bond_indices, ...]`
    if `make_distances=False` .
    The optional tensor :obj:`multiplicities` is a node-like feature tensor with a single value that gives
    the multiplicity for each node.
    The optional tensor :obj:`symmops` is an edge-like feature tensor with a matrix of shape `(4, 4)` for each edge
    that defines the symmetry operation.

    %s

    **Model outputs**:
    The standard output template:

    %s

    Args:
        inputs (list): List of dictionaries unpacked in :obj:`tf.keras.layers.Input`. Order must match model definition.
        input_tensor_type (str): Input type of graph tensor. Default is "padded".
        input_embedding (dict): Deprecated in favour of input_node_embedding etc.
        cast_disjoint_kwargs (dict): Dictionary of arguments for casting layers.
        input_node_embedding (dict): Dictionary of embedding arguments for nodes unpacked in :obj:`Embedding` layers.
        make_distances (bool): Whether input is distance or coordinates at in place of edges.
        expand_distance (bool): If the edge input are actual edges or node coordinates instead that are expanded to
            form edges with a gauss distance basis given edge indices. Expansion uses `gauss_args`.
        representation (str): The representation of unit cell. Can be either `None`, 'asu' or 'unit'. Default is 'unit'.
        conv_layer_args (dict):
        depth (int): Number of graph embedding units or depth of the network.
        verbose (int): Level of verbosity.
        name (str): Name of the model.
        gauss_args (dict): Dictionary of layer arguments unpacked in :obj:`GaussBasisLayer` layer.
        node_pooling_args (dict): Dictionary of layer arguments unpacked in :obj:`PoolingNodes` layers.
        output_embedding (str): Main embedding task for graph network. Either "node", "edge" or "graph".
        output_mlp (dict): Dictionary of layer arguments unpacked in the final classification :obj:`MLP` layer block.
            Defines number of model outputs and activation.
        output_scaling (dict): Dictionary of layer arguments unpacked in scaling layers. Default is None.
        output_tensor_type (str): Output type of graph tensors such as nodes or edges. Default is "padded".
        output_to_tensor (bool): Deprecated in favour of `output_tensor_type` .

    Returns:
        :obj:`keras.models.Model`
    """
    # Make input
    model_inputs = [Input(**x) for x in inputs]

    d_in = template_cast_list_input(
        model_inputs,
        input_tensor_type=input_tensor_type,
        cast_disjoint_kwargs=cast_disjoint_kwargs,
        mask_assignment=[0, 0 if make_distances else 1, 1, 1, None] + ([0, 1] if representation == "asu" else []),
        index_assignment=[None, None, 0, None, None] + ([None, None] if representation == "asu" else [])
    )

    if representation == "asu":
        n, x, djx, img, lattice, m, sym, batch_id_node, batch_id_edge, node_id, edge_id, count_nodes, count_edges = d_in
    else:
        n, x, djx, img, lattice, batch_id_node, batch_id_edge, node_id, edge_id, count_nodes, count_edges = d_in
        m, sym = None, None

    # Wrapp disjoint model
    out = model_disjoint_crystal(
        [n, m, x, sym, djx, img, lattice, batch_id_node, batch_id_edge, count_nodes, count_edges],
        use_node_embedding=("int" in inputs[0]['dtype']) if input_node_embedding is not None else False,
        representation=representation,
        output_embedding=output_embedding,
        input_node_embedding=input_node_embedding,
        expand_distance=expand_distance,
        conv_layer_args=conv_layer_args,
        make_distances=make_distances,
        depth=depth,
        gauss_args=gauss_args,
        node_pooling_args=node_pooling_args,
        output_mlp=output_mlp
    )

    if output_scaling is not None:
        scaler = get_scaler(output_scaling["name"])(**output_scaling)
        if scaler.extensive:
            # Node information must be numbers, or we need an additional input.
            out = scaler([out, n, batch_id_node])
        else:
            out = scaler(out)

    # Output embedding choice
    out = template_cast_output(
        [out, batch_id_node, batch_id_edge, node_id, edge_id, count_nodes, count_edges],
        output_embedding=output_embedding,
        output_tensor_type=output_tensor_type,
        input_tensor_type=input_tensor_type,
        cast_disjoint_kwargs=cast_disjoint_kwargs
    )

    model = ks.models.Model(inputs=model_inputs, outputs=out, name=name)

    model.__kgcnn_model_version__ = __model_version__

    if output_scaling is not None:
        def set_scale(*args, **kwargs):
            scaler.set_scale(*args, **kwargs)

        setattr(model, "set_scale", set_scale)

    return model


make_crystal_model.__doc__ = make_crystal_model.__doc__ % (template_cast_list_input_docs, template_cast_output_docs)



import tensorflow as tf
from kgcnn import __safe_scatter_max_min_to_zero__ as global_safe_scatter_max_min_to_zero


def scatter_reduce_sum(indices, values, shape):
    indices = tf.expand_dims(indices, axis=1)
    return tf.scatter_nd(indices, values, tf.cast(shape, dtype="int64"))


def scatter_reduce_min(indices, values, shape):
    indices = tf.expand_dims(indices, axis=1)
    target = tf.cast(tf.fill(shape, values.dtype.max), dtype=values.dtype)
    out = tf.tensor_scatter_nd_min(target, indices, values)
    if global_safe_scatter_max_min_to_zero:
        has_scattered = tf.scatter_nd(indices, tf.ones_like(values, dtype="bool"), tf.cast(shape, dtype="int64"))
        out = tf.where(has_scattered, out, tf.zeros_like(out))
    return out


def scatter_reduce_max(indices, values, shape):
    indices = tf.expand_dims(indices, axis=1)
    target = tf.cast(tf.fill(shape, values.dtype.min), dtype=values.dtype)
    out = tf.tensor_scatter_nd_max(target, indices, values)
    if global_safe_scatter_max_min_to_zero:
        has_scattered = tf.scatter_nd(indices, tf.ones_like(values, dtype="bool"), tf.cast(shape, dtype="int64"))
        out = tf.where(has_scattered, out, tf.zeros_like(out))
    return out


def scatter_reduce_mean(indices, values, shape):
    indices = tf.expand_dims(indices, axis=1)
    counts = tf.scatter_nd(indices, tf.ones_like(values), shape)
    return tf.math.divide_no_nan(tf.scatter_nd(indices, values, shape), counts)


def scatter_reduce_softmax(indices, values, shape, normalize: bool = False):
    indices_scatter = tf.expand_dims(indices, axis=1)

    if normalize:
        zeros_min = tf.zeros(shape, dtype=values.dtype)  # Zero is okay here
        data_segment_max = tf.tensor_scatter_nd_max(zeros_min, indices_scatter, values)
        data_max = tf.gather(data_segment_max, indices, axis=0)
        values = values - data_max

    values_exp = tf.math.exp(values)
    values_exp_sum = tf.scatter_nd(indices_scatter, values_exp, tf.cast(shape, dtype="int64"))
    values_exp_sum = tf.gather(values_exp_sum, indices, axis=0)
    return values_exp / values_exp_sum


def repeat_static_length(x, repeats, axis=None, total_repeat_length: int = None):
    return tf.repeat(x, repeats=repeats, axis=axis)


def decompose_ragged_tensor(x, batch_dtype="int64"):
    row_ids = tf.cast(x.value_rowids(), dtype=batch_dtype)
    row_lengths = tf.cast(x.row_lengths(), dtype=batch_dtype)
    row_starts = tf.cast(x.row_starts(), dtype=batch_dtype)
    in_ids = tf.range(tf.shape(row_ids)[0], dtype=batch_dtype)
    in_ids = in_ids - tf.repeat(row_starts, row_lengths, axis=0)
    # in_ids = tf.ragged.range(row_lengths).values
    return x.values, row_ids, in_ids, row_lengths


def norm(x, ord='fro', axis=None, keepdims=False):
    return tf.norm(x, ord=ord, axis=axis, keepdims=keepdims)


def cross(x1, x2):
    return tf.linalg.cross(x1, x2)















from ._model import MEGAN, shifted_sigmoid, ExplanationSparsityRegularization
from ._make import make_model


__all__ = [
    "make_model",
    "MEGAN",
    "ExplanationSparsityRegularization",
    "shifted_sigmoid"
]



import numpy as np
from keras import ops
from kgcnn.utils.tests import TestCase
from kgcnn.layers.aggr import AggregateLocalEdges, AggregateLocalEdgesAttention


class TestAggregateLocalEdges(TestCase):
    node_attr = np.array([[0.0, 0.0], [0.0, 1.0], [1.0, 0.0], [1.0, 1.0]])
    edge_attr = np.array([[0.0, 0.0, 0.0], [0.0, 0.0, 1.0], [0.0, 1.0, 0.0], [1.0, 1.0, 1.0],
                          [1.0, 0.0, 0.0], [1.0, 0.0, 1.0], [1.0, 1.0, 0.0], [1.0, 1.0, 1.0]])
    edge_index = np.array([[0, 0, 1, 1, 2, 2, 3, 3],
                           [0, 1, 0, 1, 2, 3, 2, 3]], dtype="int64")
    batch = np.array([0, 0, 1, 1])

    def test_correctness(self):
        layer = AggregateLocalEdges(pooling_index=1)
        nodes_aggr = layer([self.node_attr, self.edge_attr, ops.cast(self.edge_index, dtype="int64")])
        expected_output = np.array([[0., 1., 0.], [1., 1., 2.], [2., 1., 0.], [2., 1., 2.]])
        self.assertAllClose(nodes_aggr, expected_output)

    def test_correctness_mean(self):
        layer = AggregateLocalEdges(pooling_method="mean", pooling_index=0)
        nodes_aggr = layer([self.node_attr, self.edge_attr, ops.cast(self.edge_index, dtype="int64")])
        expected_output = np.array([[0., 0., 0.5], [0.5, 1., 0.5], [1., 0., 0.5], [1., 1., 0.5]])
        self.assertAllClose(nodes_aggr, expected_output)


class TestAggregateLocalEdgesAttention(TestCase):
    node_attr = np.array([[0.0, 0.0], [0.0, 1.0], [1.0, 0.0], [1.0, 1.0]])
    edge_attr = np.array([[0.0, 0.0, 0.0], [0.0, 0.0, 1.0], [0.0, 1.0, 0.0], [1.0, 1.0, 1.0],
                          [1.0, 0.0, 0.0], [1.0, 0.0, 1.0], [1.0, 1.0, 0.0], [1.0, 1.0, 1.0]])
    edge_att = np.array([[0.5], [2.0], [0.0], [1.0], [1.0], [1.0], [10.0], [1.0]])
    edge_index = np.array([[0, 1, 0, 1, 2, 3, 2, 3], [0, 0, 1, 1, 2, 2, 3, 3]], dtype="int64")
    batch = np.array([0, 0, 1, 1])

    def test_correctness(self):
        layer = AggregateLocalEdgesAttention(pooling_index=1)
        nodes_aggr = layer([self.node_attr, self.edge_attr, self.edge_att, ops.cast(self.edge_index, dtype="int64")])

        expected_output = [[0.0000000e+00, 0.0000000e+00, 8.1757444e-01],
                           [7.3105860e-01, 1.0000000e+00, 7.3105860e-01],
                           [1.0000000e+00, 0.0000000e+00, 5.0000000e-01],
                           [1.0000000e+00, 1.0000000e+00, 1.2339458e-04]]

        self.assertAllClose(nodes_aggr, expected_output)


if __name__ == "__main__":
    TestAggregateLocalEdges().test_correctness()
    TestAggregateLocalEdges().test_correctness_mean()
    TestAggregateLocalEdgesAttention().test_correctness()
    print("Tests passed.")



import keras as ks
from keras.layers import Activation, Layer
from keras import ops


class RelationalDense(Layer):
    r"""Relational :obj:`Dense` for node or edge attributes, embeddings or features.

    A :obj:`RelationalDense` layer computes a densely-connected NN layer, i.e. a linear transformation of the input
    :math:`\mathbf{x}` with the kernel weights matrix :math:`\mathbf{W}_r` and bias :math:`\mathbf{b}`
    plus (possibly non-linear) activation function :math:`\sigma` for each type of relation :math:`r` that underlies
    the feature or embedding. Examples are different edge or node types such as chemical bonds and atomic species.

    .. math::

        \mathbf{x}'_r = \sigma (\mathbf{x}_r \mathbf{W}_r + \mathbf{b})

    This has been proposed by `Schlichtkrull et al. (2017) <https://arxiv.org/abs/1703.06103>`__ for graph networks.
    Additionally, there are a set of regularization schemes to improve performance and reduce the number of learnable
    parameters proposed by `Schlichtkrull et al. (2017) <https://arxiv.org/abs/1703.06103>`__ .
    Here, the following is implemented: basis-, block-diagonal-decomposition.
    With the basis decom-position, each :math:`\mathbf{W}_r` is defined as follows:

     .. math::

        \mathbf{W}_r = \sum_{b=1}^{B} a_{rb}\; \mathbf{V}_b

    i.e. as a linear combination of basis transformations :math:`V_b \in \mathbb{R}^{d' \times d}` with coefficients
    :math:`a_{rb}` such that only the coefficients depend on :math:`r`.
    In the block-diagonal decomposition, let each :math:`W_r` be defined through the direct sum over a set of
    low-dimensional matrices:

    .. math::

        \mathbf{W}_r = \otimes_{b=1}^{B} \mathbf{Q}_{br}

    Thereby, :math:`W_r` are block-diagonal matrices: :math:`diag(Q_{1r} , \dots , Q_{Br})` with
    :math:`Q_{br} \in \mathbb{R}^{(d'/B)\times(d/B)}`.
    Usage:

    .. code-block:: python

        from keras import ops
        from kgcnn.layers.relational import RelationalDense
        f = ops.convert_to_tensor([[0., 1.], [2., 2.]])
        r = ops.convert_to_tensor([1, 2])
        layer = RelationalDense(6, num_relations=5, num_bases=3, num_blocks=2)
        out = layer([f, r])
    """

    def __init__(self,
                 units: int,
                 num_relations: int,
                 num_bases: int = None,
                 num_blocks: int = None,
                 activation=None,
                 use_bias: bool = True,
                 kernel_initializer='glorot_uniform',
                 bias_initializer='zeros',
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 **kwargs):
        r"""Initialize layer similar to :obj:`ks.layers.Dense`.

         Args:
            units: Positive integer, dimensionality of the output space.
            num_relations: Number of relations expected to construct weights.
            num_bases: Number of kernel basis functions to construct relations. Default is None.
            num_blocks: Number of block-matrices to get for parameter reduction. Default is None.
            activation: Activation function to use.
                If you don't specify anything, no activation is applied
                (ie. "linear" activation: `a(x) = x`).
            use_bias: Boolean, whether the layer uses a bias vector.
            kernel_initializer: Initializer for the `kernel` weights matrix.
            bias_initializer: Initializer for the bias vector.
            kernel_regularizer: Regularizer function applied to
                the `kernel` weights matrix.
            bias_regularizer: Regularizer function applied to the bias vector.
            activity_regularizer: Regularizer function applied to
                the output of the layer (its "activation").
            kernel_constraint: Constraint function applied to
                the `kernel` weights matrix.
            bias_constraint: Constraint function applied to the bias vector.
        """
        super(RelationalDense, self).__init__(**kwargs)
        self.units = units
        self.use_bias = use_bias
        self.num_relations = num_relations
        self.num_bases = num_bases
        self.num_blocks = num_blocks
        self.kernel_initializer = ks.initializers.get(kernel_initializer)
        self.bias_initializer = ks.initializers.get(bias_initializer)
        self.kernel_regularizer = ks.regularizers.get(kernel_regularizer)
        self.bias_regularizer = ks.regularizers.get(bias_regularizer)
        self.kernel_constraint = ks.constraints.get(kernel_constraint)
        self.bias_constraint = ks.constraints.get(bias_constraint)
        self._layer_activation = Activation(activation=activation, activity_regularizer=activity_regularizer)

    def build(self, input_shape):
        """Build layer, i.e. check input and construct weights for this layer.

        Args:
            input_shape: Shape of the input.

        Returns:
            None.
        """
        assert len(input_shape) == 2, "`RelationalDense` layer requires feature plus relation information."

        feature_shape = input_shape[0]
        relation_shape = input_shape[1]
        last_dim = feature_shape[-1]
        if last_dim is None:
            raise ValueError(
                "The last dimension of the inputs to a Dense layer "
                "should be defined. Found None. "
                f"Full input shape received: {feature_shape}"
            )
        assert len(feature_shape) - 1 == len(relation_shape), "Relations must be without feature dimension."

        # Find kernel shape.
        if self.num_bases is not None:
            num_multi_kernels = self.num_bases
        else:
            num_multi_kernels = self.num_relations
        if self.num_blocks is not None:
            assert (last_dim % self.num_blocks == 0
                    and self.units % self.num_blocks == 0), "`num_blocks` must divide in- and output dimension."
            in_kernel_dim = int(last_dim / self.num_blocks)
            out_kernel_dim = int(self.units / self.num_blocks)
        else:
            in_kernel_dim = last_dim
            out_kernel_dim = self.units
        kernel_shape = [num_multi_kernels, in_kernel_dim, out_kernel_dim]
        if self.num_blocks is not None:
            kernel_shape = [num_multi_kernels, self.num_blocks, in_kernel_dim, out_kernel_dim]
        # Make kernel
        self.kernel = self.add_weight(
            name="kernel",
            shape=tuple(kernel_shape),
            initializer=self._multi_kernel_initializer,
            regularizer=self.kernel_regularizer,
            constraint=self.kernel_constraint,
            dtype=self.dtype, trainable=True,
        )
        if self.num_bases is not None:
            self.lin_bases = self.add_weight(
                name="lin_bases",
                shape=tuple([self.num_relations, self.num_bases]),
                initializer=self.kernel_initializer,
                regularizer=self.kernel_regularizer,
                constraint=self.kernel_constraint,
                dtype=self.dtype, trainable=True,
            )
        else:
            self.lin_bases = None

        if self.use_bias:
            self.bias = self.add_weight(
                name="bias",
                shape=(self.units, ),
                initializer=self.bias_initializer,
                regularizer=self.bias_regularizer,
                constraint=self.bias_constraint,
                dtype=self.dtype, trainable=True,
            )
        else:
            self.bias = None

        self.built = True

    def compute_output_shape(self, input_shape):
        """Compute output shape."""
        assert len(input_shape) == 2, "`RelationalDense` requires '[features, relations]'."
        output_shape = list(input_shape[0])
        output_shape[-1] = self.units
        return tuple(output_shape)

    def _multi_kernel_initializer(self, shape, dtype=None, **kwargs):
        """Initialize multiple relational kernels.

        Args:
            shape: Shape of multi-kernel tensor.
            dtype: Optional dtype of the tensor.
            kwargs: Additional keyword arguments.

        Returns:
            Tensor: Tensor for initialization.
        """
        # For blocks the ks.initialize seems to have proper behaviour.
        # Each block is treated as a convolution field.
        if len(shape) < 3:
            return self.kernel_initializer(shape, dtype=dtype, **kwargs)
        # Initialize each kernel separately.
        separate_kernels = [
            ops.expand_dims(self.kernel_initializer(shape[1:], dtype=dtype, **kwargs), axis=0) for _ in range(shape[0])]
        relational_kernel = ops.concatenate(separate_kernels, axis=0)
        return relational_kernel

    def call(self, inputs, **kwargs):
        r"""Forward pass. Here, the relation is assumed to be encoded at axis=1.

        Args:
            inputs: [features, relations]

                - features (Tensor): Feature tensor of shape `([N], F)` of type 'float'.
                - relations (Tensor): Relation tensor of shape `([N], )` of type 'int'.

        Returns:
            Tensor: Processed feature tensor. Shape is `([N], units)` of type 'float'.
        """
        features, relations = inputs

        if self.num_bases is not None:
            kernel = ops.tensordot(self.lin_bases, self.kernel, [1, 0])
        else:
            kernel = self.kernel

        if self.num_blocks is not None:
            kernel_list = ops.split(kernel, ops.shape(kernel)[1], axis=1)
            kernel_list = [ops.squeeze(x, axis=1) for x in kernel_list]
            kernel_per_feature_list = [ops.take(x, relations, axis=0) for x in kernel_list]
            features_list = ops.split(features, ops.shape(features)[-1], axis=-1)
            new_feature_list = [self.batch_dot(x, y) for x, y in zip(features_list, kernel_per_feature_list)]
            new_features = ops.concatenate(new_feature_list, axis=-1)
        else:
            kernel_per_feature = ops.take(self.kernel, relations, axis=0)
            new_features = self.batch_dot(features, kernel_per_feature)

        if self.use_bias:
            new_features = new_features + self.bias
        new_features = self._layer_activation(new_features, **kwargs)
        return new_features

    @staticmethod
    def batch_dot(x, k):
        # x as features (..., N)
        # k as kernels (..., N, M)
        return ops.sum(ops.expand_dims(x, axis=-1) * k, axis=-2, keepdims=False)

    def get_config(self):
        """Update layer config."""
        config = super().get_config()
        config.update({
            "units": self.units,
            "use_bias": self.use_bias,
            "num_relations": self.num_relations,
            "num_bases": self.num_bases,
            "num_blocks": self.num_blocks,
            "kernel_initializer": ks.initializers.serialize(self.kernel_initializer),
            "bias_initializer": ks.initializers.serialize(self.bias_initializer),
            "kernel_regularizer": ks.regularizers.serialize(self.kernel_regularizer),
            "bias_regularizer": ks.regularizers.serialize(self.bias_regularizer),
            "kernel_constraint": ks.constraints.serialize(self.kernel_constraint),
            "bias_constraint": ks.constraints.serialize(self.bias_constraint),
        })
        config_act = self._layer_activation.get_config()
        for x in ["activation", "activity_regularizer"]:
            if x in config_act.keys():
                config.update({x: config_act[x]})
        return config



def broadcast_shapes(shape1, shape2):
    """Broadcast input shapes to a unified shape.

    Convert to list for mutability.

    Args:
        shape1: A tuple or list of integers.
        shape2: A tuple or list of integers.

    Returns:
        output_shape (list of integers or `None`): The broadcasted shape.

    Example:
    >>> broadcast_shapes((5, 3), (1, 3))
    [5, 3]
    """
    shape1 = list(shape1)
    shape2 = list(shape2)
    origin_shape1 = shape1
    origin_shape2 = shape2

    if len(shape1) > len(shape2):
        shape2 = [1] * (len(shape1) - len(shape2)) + shape2
    if len(shape1) < len(shape2):
        shape1 = [1] * (len(shape2) - len(shape1)) + shape1
    output_shape = list(shape1)
    for i in range(len(shape1)):
        if shape1[i] == 1:
            output_shape[i] = shape2[i]
        elif shape1[i] is None:
            output_shape[i] = None if shape2[i] == 1 else shape2[i]
        else:
            if shape2[i] == 1 or shape2[i] is None or shape2[i] == shape1[i]:
                output_shape[i] = shape1[i]
            else:
                raise ValueError(
                    "Cannot broadcast shape, the failure dim has value "
                    f"{shape1[i]}, which cannot be broadcasted to {shape2[i]}. "
                    f"Input shapes are: {origin_shape1} and {origin_shape2}."
                )

    return output_shape


# Found in ops/array_ops and copied here for static reference
def get_positive_axis(axis, ndims, axis_name="axis", ndims_name="ndims"):
    """Validate an `axis` parameter, and normalize it to be positive.
    If `ndims` is known (i.e., not `None`), then check that `axis` is in the
    range `-ndims <= axis < ndims`, and return `axis` (if `axis >= 0`) or
    `axis + ndims` (otherwise).
    If `ndims` is not known, and `axis` is positive, then return it as-is.
    If `ndims` is not known, and `axis` is negative, then report an error.

    Args:
        axis: An integer constant
        ndims: An integer constant, or `None`
        axis_name: The name of `axis` (for error messages).
        ndims_name: The name of `ndims` (for error messages).

    Returns:
        The normalized `axis` value.

    Raises:
        ValueError: If `axis` is out-of-bounds, or if `axis` is negative and `ndims is None`.
    """
    if not isinstance(axis, int):
        raise TypeError("%s must be an int; got %s" %
                        (axis_name, type(axis).__name__))
    if ndims is not None:
        if 0 <= axis < ndims:
            return axis
        elif -ndims <= axis < 0:
            return axis + ndims
        else:
            raise ValueError("%s=%s out of bounds: expected %s<=%s<%s" % (axis_name, axis, -ndims, axis_name, ndims))
    elif axis < 0:
        raise ValueError("%s may only be negative if %s is statically known." % (axis_name, ndims_name))
    return axis



from typing import Union
import unittest
import numpy as np
from keras import ops


def compare_static_shapes(found: Union[tuple, None], expected: Union[tuple, None]):
    if found is None and expected is None:
        return True
    elif found is None and expected is not None:
        return False
    elif found is not None and expected is None:
        return True
    elif len(found) != len(expected):
        return False
    shapes_okay = []
    for f, e in zip(found, expected):
        if f is None and e is not None:
            shapes_okay.append(False)
        elif f is not None and e is None:
            shapes_okay.append(True)
        elif f is None and e is None:
            shapes_okay.append(True)
        elif f == e:
            shapes_okay.append(True)
        else:
            shapes_okay.append(False)
    return all(shapes_okay)


class TestCase(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @staticmethod
    def assertAllClose(x1, x2, atol=1e-6, rtol=1e-6, msg=None):
        if not isinstance(x1, np.ndarray):
            x1 = ops.convert_to_numpy(x1)
        if not isinstance(x2, np.ndarray):
            x2 = ops.convert_to_numpy(x2)
        np.testing.assert_allclose(x1, x2, atol=atol, rtol=rtol)

    def assertNotAllClose(self, x1, x2, atol=1e-6, rtol=1e-6, msg=None):
        try:
            self.assertAllClose(x1, x2, atol=atol, rtol=rtol, msg=msg)
        except AssertionError:
            return
        msg = msg or ""
        raise AssertionError(
            f"The two values are close at all elements. \n"
            f"{msg}.\n"
            f"Values: {x1}"
        )

    def assertAllEqual(self, x1, x2, msg=None):
        self.assertEqual(len(x1), len(x2), msg=msg)
        for e1, e2 in zip(x1, x2):
            if isinstance(e1, (list, tuple)) or isinstance(e2, (list, tuple)):
                self.assertAllEqual(e1, e2, msg=msg)
            else:
                e1 = ops.convert_to_numpy(e1)
                e2 = ops.convert_to_numpy(e2)
                self.assertEqual(e1, e2, msg=msg)

    def assertLen(self, iterable, expected_len, msg=None):
        self.assertEqual(len(iterable), expected_len, msg=msg)



import keras as ks
from kgcnn.layers.scale import get as get_scaler
from ._model import model_disjoint, model_disjoint_crystal
from kgcnn.layers.modules import Input
from kgcnn.models.casting import (template_cast_output, template_cast_list_input,
                                  template_cast_list_input_docs, template_cast_output_docs)
from kgcnn.models.utils import update_model_kwargs
from keras.backend import backend as backend_to_use

# To be updated if model is changed in a significant way.
__model_version__ = "2023-09-07"

# Supported backends
__kgcnn_model_backend_supported__ = ["tensorflow", "torch", "jax"]
if backend_to_use() not in __kgcnn_model_backend_supported__:
    raise NotImplementedError("Backend '%s' for model 'Schnet' is not supported." % backend_to_use())

# Implementation of Schnet in `keras` from paper:
# by Kristof T. Schütt, Pieter-Jan Kindermans, Huziel E. Sauceda, Stefan Chmiela,
# Alexandre Tkatchenko, Klaus-Robert Müller (2018)
# https://doi.org/10.1063/1.5019779
# https://arxiv.org/abs/1706.08566
# https://aip.scitation.org/doi/pdf/10.1063/1.5019779


model_default = {
    "name": "Schnet",
    "inputs": [
        {"shape": (None,), "name": "node_number", "dtype": "int64"},
        {"shape": (None, 3), "name": "node_coordinates", "dtype": "float32"},
        {"shape": (None, 2), "name": "edge_indices", "dtype": "int64"},
        {"shape": (), "name": "total_nodes", "dtype": "int64"},
        {"shape": (), "name": "total_edges", "dtype": "int64"}
    ],
    "input_tensor_type": "padded",
    "input_embedding": None,  # deprecated
    "cast_disjoint_kwargs": {},
    "input_node_embedding": {"input_dim": 95, "output_dim": 64},
    "make_distance": True,
    "expand_distance": True,
    "interaction_args": {
        "units": 128,
        "use_bias": True,
        "activation": {"class_name": "function", "config": "kgcnn>shifted_softplus"},
        "cfconv_pool": "scatter_sum"
    },
    "node_pooling_args": {"pooling_method": "sum"},
    "depth": 4,
    "gauss_args": {"bins": 20, "distance": 4, "offset": 0.0, "sigma": 0.4},
    "verbose": 10,
    "last_mlp": {"use_bias": [True, True], "units": [128, 64],
                 "activation": [
                     {"class_name": "function", "config": "kgcnn>shifted_softplus"},
                     {"class_name": "function", "config": "kgcnn>shifted_softplus"}
                 ]},
    "output_embedding": "graph",
    "output_to_tensor": None,  # deprecated
    "use_output_mlp": True,
    "output_tensor_type": "padded",
    "output_scaling": None,
    "output_mlp": {"use_bias": [True, True], "units": [64, 1],
                   "activation": [
                       {"class_name": "function", "config": "kgcnn>shifted_softplus"},
                       "linear"
                   ]}
}


@update_model_kwargs(model_default, update_recursive=0, deprecated=["input_embedding", "output_to_tensor"])
def make_model(inputs: list = None,
               input_tensor_type: str = None,
               cast_disjoint_kwargs: dict = None,
               input_embedding: dict = None,
               input_node_embedding: dict = None,
               make_distance: bool = None,
               expand_distance: bool = None,
               gauss_args: dict = None,
               interaction_args: dict = None,
               node_pooling_args: dict = None,
               depth: int = None,
               name: str = None,
               verbose: int = None,
               last_mlp: dict = None,
               output_embedding: str = None,
               use_output_mlp: bool = None,
               output_to_tensor: bool = None,
               output_mlp: dict = None,
               output_tensor_type: str = None,
               output_scaling: dict = None
               ):
    r"""Make `SchNet <https://arxiv.org/abs/1706.08566>`__ graph network via functional API.
    Default parameters can be found in :obj:`kgcnn.literature.Schnet.model_default` .

    **Model inputs**:
    Model uses the list template of inputs and standard output template.
    The supported inputs are  :obj:`[nodes, coordinates, edge_indices, ...]` with `make_distance` and
    with '...' indicating mask or ID tensors following the template below.
    Note that you could also supply edge features with `make_distance` to False, which would make the input
    :obj:`[nodes, edges, edge_indices, ...]` .

    %s

    **Model outputs**:
    The standard output template:

    %s

    Args:
        inputs (list): List of dictionaries unpacked in :obj:`Input`. Order must match model definition.
        input_tensor_type (str): Input type of graph tensor. Default is "padded".
        cast_disjoint_kwargs (dict): Dictionary of arguments for casting layer.
        input_embedding (dict): Deprecated in favour of input_node_embedding etc.
        input_node_embedding (dict): Dictionary of embedding arguments for nodes unpacked in :obj:`Embedding` layers.
        make_distance (bool): Whether input is distance or coordinates at in place of edges.
        expand_distance (bool): If the edge input are actual edges or node coordinates instead that are expanded to
            form edges with a gauss distance basis given edge indices. Expansion uses `gauss_args`.
        gauss_args (dict): Dictionary of layer arguments unpacked in :obj:`GaussBasisLayer` layer.
        depth (int): Number of graph embedding units or depth of the network.
        interaction_args (dict): Dictionary of layer arguments unpacked in final :obj:`SchNetInteraction` layers.
        node_pooling_args (dict): Dictionary of layer arguments unpacked in :obj:`PoolingNodes` layers.
        verbose (int): Level of verbosity.
        name (str): Name of the model.
        last_mlp (dict): Dictionary of layer arguments unpacked in last :obj:`MLP` layer before output or pooling.
        output_embedding (str): Main embedding task for graph network. Either "node", "edge" or "graph".
        use_output_mlp (bool): Whether to use the final output MLP. Possibility to skip final MLP.
        output_to_tensor (bool): Deprecated in favour of `output_tensor_type` .
        output_mlp (dict): Dictionary of layer arguments unpacked in the final classification :obj:`MLP` layer block.
            Defines number of model outputs and activation.
        output_scaling (dict): Dictionary of layer arguments unpacked in scaling layers. Default is None.
        output_tensor_type (str): Output type of graph tensors such as nodes or edges. Default is "padded".

    Returns:
        :obj:`keras.models.Model`
    """
    # Make input
    model_inputs = [Input(**x) for x in inputs]

    disjoint_inputs = template_cast_list_input(
        model_inputs, input_tensor_type=input_tensor_type,
        cast_disjoint_kwargs=cast_disjoint_kwargs,
        mask_assignment=[0, 0 if make_distance else 1, 1],
        index_assignment=[None, None, 0]
    )

    n, x, disjoint_indices, batch_id_node, batch_id_edge, node_id, edge_id, count_nodes, count_edges = disjoint_inputs

    out = model_disjoint(
        [n, x, disjoint_indices, batch_id_node, count_nodes],
        use_node_embedding=("int" in inputs[0]['dtype']) if input_node_embedding is not None else False,
        input_node_embedding=input_node_embedding,
        make_distance=make_distance, expand_distance=expand_distance, gauss_args=gauss_args,
        interaction_args=interaction_args, node_pooling_args=node_pooling_args, depth=depth,
        last_mlp=last_mlp, output_embedding=output_embedding, use_output_mlp=use_output_mlp,
        output_mlp=output_mlp)

    if output_scaling is not None:
        scaler = get_scaler(output_scaling["name"])(**output_scaling)
        if scaler.extensive:
            # Node information must be numbers, or we need an additional input.
            out = scaler([out, n, batch_id_node])
        else:
            out = scaler(out)

    # Output embedding choice
    out = template_cast_output(
        [out, batch_id_node, batch_id_edge, node_id, edge_id, count_nodes, count_edges],
        output_embedding=output_embedding, output_tensor_type=output_tensor_type,
        input_tensor_type=input_tensor_type, cast_disjoint_kwargs=cast_disjoint_kwargs,
    )

    model = ks.models.Model(inputs=model_inputs, outputs=out, name=name)

    model.__kgcnn_model_version__ = __model_version__

    if output_scaling is not None:
        def set_scale(*args, **kwargs):
            scaler.set_scale(*args, **kwargs)

        setattr(model, "set_scale", set_scale)

    return model


make_model.__doc__ = make_model.__doc__ % (template_cast_list_input_docs, template_cast_output_docs)


model_crystal_default = {
    "name": "Schnet",
    "inputs": [
        {"shape": (None,), "name": "node_number", "dtype": "int64"},
        {"shape": (None, 3), "name": "node_coordinates", "dtype": "float32"},
        {"shape": (None, 2), "name": "edge_indices", "dtype": "int64"},
        {"shape": (None, 3), "name": "edge_image", "dtype": "int64"},
        {"shape": (3, 3), "name": "graph_lattice", "dtype": "float32"},
        {"shape": (), "name": "total_nodes", "dtype": "int64"},
        {"shape": (), "name": "total_edges", "dtype": "int64"}
    ],
    "input_tensor_type": "padded",
    "input_embedding": None,  # deprecated
    "cast_disjoint_kwargs": {},
    "input_node_embedding": {"input_dim": 95, "output_dim": 64},
    "make_distance": True, "expand_distance": True,
    "interaction_args": {
        "units": 128,
        "use_bias": True,
        "activation": {"class_name": "function", "config": "kgcnn>shifted_softplus"},
        "cfconv_pool": "sum"
    },
    "node_pooling_args": {"pooling_method": "sum"},
    "depth": 4,
    "gauss_args": {"bins": 20, "distance": 4, "offset": 0.0, "sigma": 0.4},
    "verbose": 10,
    "last_mlp": {"use_bias": [True, True], "units": [128, 64],
                 "activation": [
                     {"class_name": "function", "config": "kgcnn>shifted_softplus"},
                     {"class_name": "function", "config": "kgcnn>shifted_softplus"}
                 ]},
    "output_embedding": "graph",
    "output_to_tensor": None,  # deprecated
    "use_output_mlp": True,
    "output_scaling": None,
    "output_tensor_type": "padded",
    "output_mlp": {"use_bias": [True, True], "units": [64, 1],
                   "activation": [
                       {"class_name": "function", "config": "kgcnn>shifted_softplus"},
                       "linear"
                   ]}
}


@update_model_kwargs(model_crystal_default, update_recursive=0, deprecated=["input_embedding", "output_to_tensor"])
def make_crystal_model(inputs: list = None,
                       input_tensor_type: str = None,
                       cast_disjoint_kwargs: dict = None,
                       input_embedding: dict = None,
                       input_node_embedding: dict = None,
                       make_distance: bool = None,
                       expand_distance: bool = None,
                       gauss_args: dict = None,
                       interaction_args: dict = None,
                       node_pooling_args: dict = None,
                       depth: int = None,
                       name: str = None,
                       verbose: int = None,
                       last_mlp: dict = None,
                       output_embedding: str = None,
                       use_output_mlp: bool = None,
                       output_to_tensor: bool = None,
                       output_mlp: dict = None,
                       output_scaling: dict = None,
                       output_tensor_type: str = None,
                       ):
    r"""Make `SchNet <https://arxiv.org/abs/1706.08566>`__ graph network via functional API.
    Default parameters can be found in :obj:`kgcnn.literature.Schnet.model_crystal_default`.

    **Model inputs**:
    Model uses the list template of inputs and standard output template.
    The supported inputs are  :obj:`[nodes, coordinates, edge_indices, image_translation, lattice, ...]`
    with '...' indicating mask or ID tensors following the template below.

    %s

    **Model outputs**:
    The standard output template:

    %s

    Args:
        inputs (list): List of dictionaries unpacked in :obj:`Input`. Order must match model definition.
        input_tensor_type (str): Input type of graph tensor. Default is "padded".
        cast_disjoint_kwargs (dict): Dictionary of arguments for casting layer.
        input_embedding (dict): Deprecated in favour of input_node_embedding etc.
        input_node_embedding (dict): Dictionary of embedding arguments for nodes unpacked in :obj:`Embedding` layers.
        make_distance (bool): Whether input is distance or coordinates at in place of edges.
        expand_distance (bool): If the edge input are actual edges or node coordinates instead that are expanded to
            form edges with a gauss distance basis given edge indices. Expansion uses `gauss_args`.
        gauss_args (dict): Dictionary of layer arguments unpacked in :obj:`GaussBasisLayer` layer.
        depth (int): Number of graph embedding units or depth of the network.
        interaction_args (dict): Dictionary of layer arguments unpacked in final :obj:`SchNetInteraction` layers.
        node_pooling_args (dict): Dictionary of layer arguments unpacked in :obj:`PoolingNodes` layers.
        verbose (int): Level of verbosity.
        name (str): Name of the model.
        last_mlp (dict): Dictionary of layer arguments unpacked in last :obj:`MLP` layer before output or pooling.
        output_embedding (str): Main embedding task for graph network. Either "node", "edge" or "graph".
        use_output_mlp (bool): Whether to use the final output MLP. Possibility to skip final MLP.
        output_to_tensor (bool): Deprecated in favour of `output_tensor_type` .
        output_mlp (dict): Dictionary of layer arguments unpacked in the final classification :obj:`MLP` layer block.
            Defines number of model outputs and activation.
        output_scaling (dict): Dictionary of layer arguments unpacked in scaling layers. Default is None.
        output_tensor_type (str): Output type of graph tensors such as nodes or edges. Default is "padded".

    Returns:
        :obj:`keras.models.Model`
    """
    # Make input
    model_inputs = [Input(**x) for x in inputs]

    disjoint_inputs = template_cast_list_input(
        model_inputs, input_tensor_type=input_tensor_type,
        cast_disjoint_kwargs=cast_disjoint_kwargs,
        mask_assignment=[0, 0 if make_distance else 1, 1, 1, None],
        index_assignment=[None, None, 0, None, None]
    )

    n, x, djx, img, lattice, batch_id_node, batch_id_edge, node_id, edge_id, count_nodes, count_edges = disjoint_inputs

    # Wrapp disjoint model
    out = model_disjoint_crystal(
        [n, x, djx, img, lattice, batch_id_node, batch_id_edge, count_nodes],
        use_node_embedding=("int" in inputs[0]['dtype']) if input_node_embedding is not None else False,
        input_node_embedding=input_node_embedding,
        make_distance=make_distance, expand_distance=expand_distance, gauss_args=gauss_args,
        interaction_args=interaction_args, node_pooling_args=node_pooling_args, depth=depth, last_mlp=last_mlp,
        output_embedding=output_embedding, use_output_mlp=use_output_mlp, output_mlp=output_mlp
    )

    if output_scaling is not None:
        scaler = get_scaler(output_scaling["name"])(**output_scaling)
        if scaler.extensive:
            # Node information must be numbers, or we need an additional input.
            out = scaler([out, n, batch_id_node])
        else:
            out = scaler(out)

    # Output embedding choice
    out = template_cast_output(
        [out, batch_id_node, batch_id_edge, node_id, edge_id, count_nodes, count_edges],
        output_embedding=output_embedding, output_tensor_type=output_tensor_type,
        input_tensor_type=input_tensor_type, cast_disjoint_kwargs=cast_disjoint_kwargs,
    )

    model = ks.models.Model(inputs=model_inputs, outputs=out, name=name)

    model.__kgcnn_model_version__ = __model_version__

    if output_scaling is not None:
        def set_scale(*args, **kwargs):
            scaler.set_scale(*args, **kwargs)

        setattr(model, "set_scale", set_scale)

    return model


make_crystal_model.__doc__ = make_crystal_model.__doc__ % (template_cast_list_input_docs, template_cast_output_docs)



import numpy as np
from kgcnn.utils.tests import TestCase
from keras import ops
from kgcnn.layers.gather import GatherNodes


class GatherNodesTest(TestCase):

    node_attr = np.array([[0.0, 0.0], [0.0, 1.0], [1.0, 0.0], [1.0, 1.0]])
    edge_attr = np.array([[0.0, 0.0, 0.0], [0.0, 0.0, 1.0], [0.0, 1.0, 0.0], [1.0, 1.0, 1.0],
                          [1.0, 0.0, 0.0], [1.0, 0.0, 1.0], [1.0, 1.0, 0.0], [-1.0, 1.0, 1.0]])
    edge_index = np.array([[0, 0, 1, 1, 2, 2, 3, 3], [0, 1, 0, 1, 2, 3, 2, 3]], dtype="int64")
    batch = np.array([0, 0, 1, 1])

    def test_correctness(self):

        layer = GatherNodes()
        nodes_per_edge = layer([self.node_attr, ops.cast(self.edge_index, dtype="int64")])
        expected_output = np.array([[0., 0., 0., 0., ], [0., 0., 0., 1.], [0., 1., 0., 0.], [0., 1., 0., 1.],
                                    [1., 0., 1., 0.], [1., 0., 1., 1.], [1., 1., 1., 0.], [1., 1., 1., 1.]])
        self.assertAllClose(nodes_per_edge, expected_output)


if __name__ == "__main__":

    GatherNodesTest().test_correctness()
    print("Tests passed.")


import keras as ks
from keras.layers import Layer, Add, Multiply, Concatenate, Dense
from kgcnn.layers.update import ResidualLayer
from kgcnn.layers.mlp import GraphMLP
from kgcnn.layers.gather import GatherNodes, GatherNodesOutgoing
from kgcnn.layers.aggr import AggregateLocalEdges as PoolingLocalMessages


class MXMGlobalMP(Layer):

    def __init__(self, units: int = 64, pooling_method="mean", **kwargs):
        """Initialize layer."""
        super(MXMGlobalMP, self).__init__(**kwargs)
        self.dim = units
        self.pooling_method = pooling_method
        self.h_mlp = GraphMLP(self.dim, activation="swish")
        self.res1 = ResidualLayer(self.dim)
        self.res2 = ResidualLayer(self.dim)
        self.res3 = ResidualLayer(self.dim)
        self.mlp = GraphMLP(self.dim, activation="swish")
        self.add_res = Add()

        self.x_edge_mlp = GraphMLP(self.dim, activation="swish")
        self.linear = Dense(self.dim, use_bias=False, activation="linear")

        self.gather = GatherNodes(split_indices=[0, 1], concat_axis=None)
        self.pool = PoolingLocalMessages(pooling_method=pooling_method)
        self.cat = Concatenate(axis=-1)
        self.multiply_edge = Multiply()
        self.add = Add()

    def build(self, input_shape):
        """Build layer."""
        super(MXMGlobalMP, self).build(input_shape)

    def propagate(self, edge_index, x, edge_attr, **kwargs):
        x_i, x_j = self.gather([x, edge_index])

        # Prepare message.
        x_edge = self.cat([x_i, x_j, edge_attr], **kwargs)
        x_edge = self.x_edge_mlp(x_edge, **kwargs)
        edge_attr_lin = self.linear(edge_attr, **kwargs)
        x_edge = self.multiply_edge([edge_attr_lin, x_edge])

        # Pooling here.
        x_p = self.pool([x, x_edge, edge_index])

        # Replace self loops by explicit node update here.
        x_i_p = self.add([x_p, x])

        return x_i_p

    def call(self, inputs, **kwargs):
        r"""Forward pass.

        Args:
            inputs: [nodes, edges, tensor_index]

                - nodes (Tensor): Node embeddings of shape `([N], F)`
                - edges (Tensor): Edge or message embeddings of shape `([M], F)`
                - tensor_index (Tensor): Edge indices referring to nodes of shape `(2, [M])`

        Returns:
            Tensor: Node embeddings.
        """
        h, edge_attr, edge_index = inputs

        # Keep for residual skip connections.
        res_h = h

        # Integrate the Cross Layer Mapping inside the Global Message Passing
        h = self.h_mlp(h)

        # Message Passing operation
        h = self.propagate(edge_index=edge_index, x=h, edge_attr=edge_attr, **kwargs)

        # Update function f_u
        h = self.res1(h)
        h = self.mlp(h)
        h = self.add_res([h, res_h])
        h = self.res2(h)
        h = self.res3(h)

        # Message Passing operation
        h = self.propagate(edge_index=edge_index, x=h, edge_attr=edge_attr, **kwargs)

        return h

    def get_config(self):
        config = super(MXMGlobalMP, self).get_config()
        config.update({"units": self.dim, "pooling_method": self.pooling_method})
        return config


class MXMLocalMP(Layer):

    def __init__(self, units: int = 64, output_units: int = 1, activation: str = "swish",
                 output_kernel_initializer: str = "zeros", pooling_method: str = "sum", **kwargs):
        super(MXMLocalMP, self).__init__(**kwargs)
        self.dim = units
        self.output_dim = output_units
        self.activation = activation
        self.pooling_method = pooling_method
        self.h_mlp = GraphMLP(self.dim, activation=activation)

        self.mlp_kj = GraphMLP([self.dim], activation=activation)
        self.mlp_ji_1 = GraphMLP([self.dim], activation=activation)
        self.mlp_ji_2 = GraphMLP([self.dim], activation=activation)
        self.mlp_jj = GraphMLP([self.dim], activation=activation)

        self.mlp_sbf1 = GraphMLP([self.dim, self.dim], activation=activation)
        self.mlp_sbf2 = GraphMLP([self.dim, self.dim], activation=activation)
        self.lin_rbf1 = Dense(self.dim, use_bias=False, activation="linear")
        self.lin_rbf2 = Dense(self.dim, use_bias=False, activation="linear")

        self.res1 = ResidualLayer(self.dim)
        self.res2 = ResidualLayer(self.dim)
        self.res3 = ResidualLayer(self.dim)

        self.lin_rbf_out = Dense(self.dim, use_bias=False, activation="linear")

        # Fix for kgcnn==4.0.1: removed overwrite mlp here. Should not change model but prevents unused layers.

        self.y_mlp = GraphMLP([self.dim, self.dim, self.dim], activation=activation)
        self.y_W = Dense(self.output_dim, activation="linear",
                         kernel_initializer=output_kernel_initializer)
        self.add_res = Add()

        self.gather_nodes = GatherNodes(split_indices=[0, 1], concat_axis=None)
        self.cat = Concatenate()
        self.multiply = Multiply()
        self.gather_mkj = GatherNodesOutgoing()
        self.gather_mjj = GatherNodesOutgoing()
        self.pool_mkj = PoolingLocalMessages(pooling_method=pooling_method)
        self.pool_mjj = PoolingLocalMessages(pooling_method=pooling_method)
        self.pool_h = PoolingLocalMessages(pooling_method=pooling_method)
        self.add_mji_1 = Add()
        self.add_mji_2 = Add()
        
    def build(self, input_shape):
        super(MXMLocalMP, self).build(input_shape)

    def call(self, inputs, **kwargs):
        r"""Forward pass.

        Args:
            inputs: [h, rbf, sbf1, sbf2, edge_index, angle_idx_1, angle_idx_2]

                - h (Tensor): Node embeddings of shape `([N], F)`
                - rbf (Tensor): Radial basis functions of shape `([M], F)`
                - sbf1 (Tensor): Spherical basis functions of shape `([K], F)`
                - sbf2 (Tensor): Spherical basis functions of shape `([K], F)`
                - edge_index (Tensor): Edge indices of shape `(2, [M])`
                - angle_idx_1 (Tensor): Angle 1 indices of shape `(2, [K])`
                - angle_idx_2 (Tensor): Angle 2 indices of shape `(2, [K])`

        Returns:
            Tensor: Node embeddings.
        """
        h, rbf, sbf1, sbf2, edge_index, angle_idx_1, angle_idx_2 = inputs
        res_h = h

        # Integrate the Cross Layer Mapping inside the Local Message Passing
        h = self.h_mlp(h, **kwargs)

        # Message Passing 1
        hi, hj = self.gather_nodes([h, edge_index])
        m = self.cat([hi, hj, rbf])

        m_kj = self.mlp_kj(m, **kwargs)
        w_rbf1 = self.lin_rbf1(rbf, **kwargs)
        m_kj = self.multiply([m_kj, w_rbf1])
        m_kj = self.gather_mkj([m_kj, angle_idx_1])
        sw_sbf1 = self.mlp_sbf1(sbf1, **kwargs)
        m_kj = self.multiply([m_kj, sw_sbf1])
        m_kj = self.pool_mkj([m, m_kj, angle_idx_1])

        m_ji_1 = self.mlp_ji_1(m, **kwargs)

        m = self.add_mji_1([m_ji_1, m_kj])

        # Message Passing 2 (index jj denotes j'i in the main paper)
        m_jj = self.mlp_jj(m, **kwargs)
        w_rbf2 = self.lin_rbf2(rbf, **kwargs)
        m_jj = self.multiply([m_jj, w_rbf2])
        m_jj = self.gather_mjj([m_jj, angle_idx_2])
        sw_sbf2 = self.mlp_sbf2(sbf2, **kwargs)
        m_jj = self.multiply([m_jj, sw_sbf2])
        m_jj = self.pool_mjj([m, m_jj, angle_idx_2])

        m_ji_2 = self.mlp_ji_2(m, **kwargs)

        m = self.add_mji_2([m_ji_2, m_jj])

        # Aggregation
        w_rbf = self.lin_rbf_out(rbf, **kwargs)
        m = self.multiply([w_rbf, m])
        h = self.pool_h([h, m, edge_index])

        # Update function f_u
        h = self.res1(h, **kwargs)
        h = self.h_mlp(h, **kwargs)
        h = self.add_res([h, res_h])
        h = self.res2(h, **kwargs)
        h = self.res3(h, **kwargs)

        # Output Module
        y = self.y_mlp(h, **kwargs)
        y = self.y_W(y, **kwargs)

        return h, y

    def get_config(self):
        config = super(MXMLocalMP, self).get_config()
        out_conf = self.y_W.get_config()
        config.update({"units": self.dim, "output_units": self.output_dim,
                       "activation": ks.activations.serialize(ks.activations.get(self.activation)),
                       "output_kernel_initializer": out_conf["kernel_initializer"],
                       "pooling_method": self.pooling_method})
        return config



from kgcnn.layers.update import GRUUpdate
from keras.layers import Flatten, Add
from kgcnn.layers.modules import Embedding, ExpandDims, SqueezeDims
from kgcnn.layers.pooling import PoolingNodesAttentive
from keras.layers import Dense, Dropout, Concatenate, Attention
from kgcnn.layers.mlp import MLP, GraphMLP
from ._layers import AttentiveHeadFP_


def model_disjoint(
        inputs,
        use_node_embedding: bool = None,
        use_edge_embedding: bool = None,
        input_node_embedding: dict = None,
        input_edge_embedding: dict = None,
        attention_args=None,
        dropout=None,
        depthato=None,
        depthmol=None,
        output_embedding=None,
        output_mlp=None,
        pooling_gat_nodes_args=None
):
    # Model implementation with disjoint representation.
    n, ed, edi, batch_id_node, count_nodes = inputs

    # Embedding, if no feature dimension
    if use_node_embedding:
        n = Embedding(**input_node_embedding)(n)
    if use_edge_embedding:
        ed = Embedding(**input_edge_embedding)(ed)

    # Model
    nk = Dense(units=attention_args['units'])(n)
    ck = AttentiveHeadFP_(use_edge_features=True, **attention_args)([nk, ed, edi])
    nk = GRUUpdate(units=attention_args['units'])([nk, ck])
    nk = Dropout(rate=dropout)(nk)  # adding dropout to the first code not in the original AttFP code ?
    list_emb = [nk]  # "aka r1"
    for i in range(1, depthato):
        ck = AttentiveHeadFP_(**attention_args)([nk, ed, edi])
        nk = GRUUpdate(units=attention_args['units'])([nk, ck])
        nk = Dropout(rate=dropout)(nk)
        list_emb.append(nk)

    # we store representation of each atomic nodes (at r1,r2,...)

    if output_embedding == 'graph':
        # we apply a super node to each atomic node representation and concate them
        out = [
            # Tensor output.
            PoolingNodesAttentive(units=attention_args['units'], depth=depthmol)([count_nodes, ni, batch_id_node]) for
            ni in list_emb
        ]
        out = [ExpandDims(axis=1)(x) for x in out]
        out = Concatenate(axis=1)(out)
        # we compute the weigthed scaled self-attention of the super nodes
        at = Attention(dropout=dropout, use_scale=True, score_mode="dot")([out, out])
        # we apply the dot product
        out = at * out
        out = Flatten()(out)
        # in the paper this is only one dense layer to the target ... very simple
        out = MLP(**output_mlp)(out)

    elif output_embedding == 'node':
        n = Add()(list_emb)
        out = GraphMLP(**output_mlp)([n, batch_id_node, count_nodes])
    else:
        raise ValueError("Unsupported graph embedding for mode `MoGAT` .")

    return out



import keras as ks
from kgcnn.layers.scale import get as get_scaler
from kgcnn.models.utils import update_model_kwargs
from kgcnn.models.casting import (template_cast_output, template_cast_list_input,
                                  template_cast_list_input_docs, template_cast_output_docs)
from keras.backend import backend as backend_to_use
from kgcnn.layers.modules import Input
from ._model import model_disjoint

# Keep track of model version from commit date in literature.
# To be updated if model is changed in a significant way.
__model_version__ = "2023-10-23"

# Supported backends
__kgcnn_model_backend_supported__ = ["tensorflow", "torch", "jax"]

if backend_to_use() not in __kgcnn_model_backend_supported__:
    raise NotImplementedError("Backend '%s' for model 'DGIN' is not supported." % backend_to_use())

# Implementation of DGIN in `keras` from paper:
# Analyzing Learned Molecular Representations for Property Prediction
# by Oliver Wieder, Mélaine Kuenemann, Marcus Wieder, Thomas Seidel,
# Christophe Meyer, Sharon D Bryant and Thierry Langer
# https://pubmed.ncbi.nlm.nih.gov/34684766/

model_default = {
    "name": "DGIN",
    "inputs": [
        {"shape": (None,), "name": "node_number", "dtype": "int64"},
        {"shape": (None,), "name": "edge_number", "dtype": "int64"},
        {"shape": (None, 2), "name": "edge_indices", "dtype": "int64"},
        {"shape": (None, 1), "name": "edge_indices_reverse", "dtype": "int64"},
        {"shape": (), "name": "total_nodes", "dtype": "int64"},
        {"shape": (), "name": "total_edges", "dtype": "int64"},
        {"shape": (), "name": "total_reverse", "dtype": "int64"}
    ],
    "input_tensor_type": "padded",
    "cast_disjoint_kwargs": {},
    "input_embedding": None,  # deprecated
    "input_node_embedding": {"input_dim": 95, "output_dim": 64},
    "input_edge_embedding": {"input_dim": 5, "output_dim": 64},
    "input_graph_embedding": {"input_dim": 100, "output_dim": 64},
    "gin_mlp": {"units": [64, 64], "use_bias": True, "activation": ["relu", "linear"],
                "use_normalization": True, "normalization_technique": "graph_batch"},
    "gin_args": {},
    "last_mlp": {"use_bias": [True, True], "units": [64, 64],
                 "activation": ["relu", "relu"]},
    "pooling_args": {"pooling_method": "sum"},
    "use_graph_state": False,
    "edge_initialize": {"units": 128, "use_bias": True, "activation": "relu"},
    "edge_dense": {"units": 128, "use_bias": True, "activation": "linear"},
    "node_dense": {"units": 128, "use_bias": True, "activation": "linear"},
    "edge_activation": {"activation": "relu"},
    "verbose": 10,
    "depthDMPNN": 4,
    "depthGIN": 4,
    "dropoutDMPNN": {"rate": 0.15},
    "dropoutGIN": {"rate": 0.15},
    "output_embedding": "graph",
    "node_pooling_kwargs": {"pooling_method": "mean"},
    "output_to_tensor": None,  # deprecated
    "output_tensor_type": "padded",
    "output_mlp": {"use_bias": True, "units": 1,
                   "activation": "linear"},
    "output_scaling": None,
}


@update_model_kwargs(model_default, update_recursive=0, deprecated=["input_embedding", "output_to_tensor"])
def make_model(name: str = None,
               inputs: list = None,
               input_tensor_type: str = None,
               cast_disjoint_kwargs: dict = None,
               input_embedding: dict = None,
               input_node_embedding: dict = None,
               input_edge_embedding: dict = None,
               input_graph_embedding: dict = None,
               pooling_args: dict = None,
               edge_initialize: dict = None,
               edge_dense: dict = None,
               node_dense: dict = None,
               edge_activation: dict = None,
               dropoutDMPNN: dict = None,
               dropoutGIN: dict = None,
               depthDMPNN: int = None,
               depthGIN: int = None,
               gin_args: dict = None,
               gin_mlp: dict = None,
               last_mlp: dict = None,
               verbose: int = None,
               node_pooling_kwargs: dict = None,
               use_graph_state: bool = False,
               output_embedding: str = None,
               output_to_tensor: bool = None,
               output_tensor_type: str = None,
               output_mlp: dict = None,
               output_scaling: dict = None
               ):
    r"""Make `DGIN <https://pubmed.ncbi.nlm.nih.gov/34684766/>`__ graph network via functional API.
    Default parameters can be found in :obj:`kgcnn.literature.DGIN.model_default` .

    **Model inputs**:
    Model uses the list template of inputs and standard output template.
    The supported inputs are  :obj:`[nodes, edges, edge_indices, reverse_indices, (graph_state), ...]`
    with '...' indicating mask or id tensors following the template below.
    Here, reverse indices are in place of angle indices and refer to edges. The graph state is optional and controlled
    by `use_graph_state` parameter.

    %s

    **Model outputs**:
    The standard output template:

    %s

    Args:
        name (str): Name of the model. Should be "DGIN".
        inputs (list): List of dictionaries unpacked in :obj:`Input`. Order must match model definition.
        input_tensor_type (str): Input type of graph tensor. Default is "padded".
        cast_disjoint_kwargs (dict): Dictionary of arguments for casting layers if used.
        input_embedding (dict): Deprecated in favour of input_node_embedding etc.
        input_node_embedding (dict): Dictionary of arguments for nodes unpacked in :obj:`Embedding` layers.
        input_edge_embedding (dict): Dictionary of arguments for edge unpacked in :obj:`Embedding` layers.
        input_graph_embedding (dict): Dictionary of arguments for edge unpacked in :obj:`Embedding` layers.
        pooling_args (dict): Dictionary of layer arguments unpacked in :obj:`AggregateLocalEdges` layers.
        edge_initialize (dict): Dictionary of layer arguments unpacked in :obj:`Dense` layer for first edge embedding.
        edge_dense (dict): Dictionary of layer arguments unpacked in :obj:`Dense` layer for edge embedding.
        node_dense (dict): Dictionary of layer arguments unpacked in :obj:`Dense` layer for node embedding.
        edge_activation (dict): Edge Activation after skip connection.
        depthDMPNN (int): Number of graph embedding units or depth of the DMPNN subnetwork.
        depthGIN (int): Number of graph embedding units or depth of the GIN subnetwork.
        dropoutDMPNN (dict): Dictionary of layer arguments unpacked in :obj:`Dropout`.
        dropoutGIN (float): dropout rate.
        gin_args (dict): Kwargs unpacked in :obj:`GIN_D` convolutional unit.
        gin_mlp (dict): Kwargs unpacked in :obj:`MLP` for GIN layer.
        last_mlp (dict): Kwargs unpacked in last :obj:`MLP` .
        verbose (int): Level for print information.
        use_graph_state (bool): Whether to use graph state information. Default is False.
        node_pooling_kwargs (dict): Dictionary of layer arguments unpacked in :obj:`PoolingNodes` layer.
        output_embedding (str): Main embedding task for graph network. Either "node", "edge" or "graph".
        output_to_tensor (bool): WDeprecated in favour of `output_tensor_type` .
        output_tensor_type (str): Output type of graph tensors such as nodes or edges. Default is "padded".
        output_mlp (dict): Dictionary of layer arguments unpacked in the final classification :obj:`MLP` layer block.
            Defines number of model outputs and activation.
        output_scaling (dict): Kwargs for scaling layer, if scaling layer is to be used.

    Returns:
        :obj:`keras.models.Model`
    """
    # Make input
    model_inputs = [Input(**x) for x in inputs]

    di = template_cast_list_input(
        model_inputs, input_tensor_type=input_tensor_type,
        cast_disjoint_kwargs=cast_disjoint_kwargs,
        index_assignment=[None, None, 0, 2] + ([None] if use_graph_state else []),
        mask_assignment=[0, 1, 1, 2] + ([None] if use_graph_state else [])
    )

    if use_graph_state:
        n, ed, edi, e_pairs, gs, batch_id_node, batch_id_edge, _, node_id, edge_id, _, count_nodes, count_edges, _ = di
    else:
        n, ed, edi, e_pairs, batch_id_node, batch_id_edge, _, node_id, edge_id, _, count_nodes, count_edges, _ = di
        gs = None

    # Wrapping disjoint model.
    out = model_disjoint(
        [n, ed, edi, batch_id_node, e_pairs, count_nodes, gs],
        use_node_embedding=("int" in inputs[0]['dtype']) if input_node_embedding is not None else False,
        use_edge_embedding=("int" in inputs[1]['dtype']) if input_edge_embedding is not None else False,
        use_graph_embedding=False if not use_graph_state else (
                "int" in inputs[4]['dtype']) if input_graph_embedding is not None else False,
        use_graph_state=use_graph_state,
        input_node_embedding=input_node_embedding,
        input_edge_embedding=input_edge_embedding,
        input_graph_embedding=input_graph_embedding,
        edge_initialize=edge_initialize,
        edge_activation=edge_activation,
        edge_dense=edge_dense,
        depthDMPNN=depthDMPNN,
        dropoutDMPNN=dropoutDMPNN,
        pooling_args=pooling_args,
        gin_mlp=gin_mlp,
        depthGIN=depthGIN,
        gin_args=gin_args,
        output_embedding=output_embedding,
        node_pooling_kwargs=node_pooling_kwargs,
        last_mlp=last_mlp,
        dropoutGIN=dropoutGIN,
        output_mlp=output_mlp
    )

    if output_scaling is not None:
        scaler = get_scaler(output_scaling["name"])(**output_scaling)
        out = scaler(out)

    # Output embedding choice
    out = template_cast_output(
        [out, batch_id_node, batch_id_edge, node_id, edge_id, count_nodes, count_edges],
        output_embedding=output_embedding, output_tensor_type=output_tensor_type,
        input_tensor_type=input_tensor_type, cast_disjoint_kwargs=cast_disjoint_kwargs
    )

    model = ks.models.Model(inputs=model_inputs, outputs=out, name=name)
    model.__kgcnn_model_version__ = __model_version__

    if output_scaling is not None:
        def set_scale(*args, **kwargs):
            scaler.set_scale(*args, **kwargs)

        setattr(model, "set_scale", set_scale)

    return model


make_model.__doc__ = make_model.__doc__ % (template_cast_list_input_docs, template_cast_output_docs)
